/* CRS (Channelflow with options of Rotation and Stratification), 
   revised Channelflow-1.1.2 (John F. Gibson) code including rotation and stratification
   Copyright (C) 2007-2008 Hao Lu

   Engine Research Center
   1500 Engineering Dr. #1011
   Madison, WI 53706

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <iomanip>
#include "dns.h"
#include "flowfield.h"
#include "time.h"

using namespace std;
using namespace channelflow;

int main(int argc, char *argv[]) {

  int ioutput=0;
  
  if (argc > 1) {
    ioutput = atoi(argv[1]);
  }

  bool saveData = true;
  if (ioutput < 0) saveData = false;
  ioutput = abs(ioutput);
  
  bool plot3d = false;
  if (ioutput == 1 || ioutput == 3 || ioutput == 4 || ioutput == 6) plot3d = true;

  bool tecplot = false;
  if (ioutput == 2 || ioutput == 3 || ioutput == 5 || ioutput == 6) tecplot = true;

  bool UPlus = false;
  if (ioutput == 4 || ioutput == 5 || ioutput == 6 || ioutput == 7) UPlus = true;

  cout << endl;
  cout << "+----------------------------------------------------------------+" << endl;
  cout << "| CRS (Channelflow with options of Rotation and Stratification), |" << endl;
  cout << "| Copyright (C) 2007-2008 Hao Lu                                 |" << endl;
  cout << "| This program comes with ABSOLUTELY NO WARRANTY.                |" << endl;
  cout << "| This is free software, and you are welcome to redistribute it  |" << endl;
  cout << "| under certain conditions.                                      |" << endl;
  cout << "+----------------------------------------------------------------+" << endl;
    
  // Define gridsize
  const int Nx=64;
  const int Ny=65;
  const int Nz=64;
  
  // flow parameters
  Real nu_;
  Real dtmax_;
  Real CFLmin_;
  Real CFLmax_;
  Real nRac_;
  Real T0_;
  Real T1_;
  Real dT_;
  Real Lx_;
  Real Ly_;
  Real Lz_;
  
  // Define DNS parameters
  DNSFlags flags; 

  if (readParameter(&nu_, &dtmax_, &CFLmin_, &CFLmax_, &nRac_,
    &T0_, &T1_, &dT_, &Lx_, &Ly_, &Lz_, &flags) != 1) {
  cout << "error: read parameter file 'condition.param'" << endl;
  exit(0);
  }

  // Define box size
  const Real a= -Ly_/2.0;
  const Real b=  Ly_/2.0;
  const Real Lx= Lx_;
  const Real Lz= Lz_;
  
  // zero mean, however, the solver requires it
  ChebyCoeff U(Ny,a,b,Physical);
  for (int ny=0; ny<Ny; ++ny) {
  U[ny] = 0.0;
  }
  
  // Define flow parameters
  const Real nu = nu_;

  // Define integration parameters
  const Real dtmin = 0.00001;
  const Real dtmax = dtmax_;
  const Real dtini = 0.003;
  const Real CFLmin = CFLmin_;
  const Real CFLmax = CFLmax_;
  const Real dT  = dT_;   // plot interval
  const Real T0  = T0_;   // start time
  const Real T1  = T1_;   // stop time

  Real t;

  // Define size and smoothness of initial disturbance
  Real spectralDecay;
  Real magnitude;
  spectralDecay = 0.4;   
  magnitude = 0.000001;

  // Construct data fields: 3d velocity, 1d pressure and 1d scalar
  cout << "building velocity and pressure fields..." << flush;
  FlowField u(Nx,Ny,Nz,3,Lx,Lz,a,b); u.setToZero();
  FlowField q(Nx,Ny,Nz,1,Lx,Lz,a,b); q.setToZero();
  FlowField s(Nx,Ny,Nz,1,Lx,Lz,a,b); s.setToZero();
  FlowField LRi(1,Ny,1,1,Lx,Lz,a,b); LRi.setToZero();
  FlowField MProfile(1,Ny,1,4,Lx,Lz,a,b); MProfile.setToZero();
  cout << "done" << endl;

//  if (fabs(flags.Omega) < 0.000001 && (fabs(flags.Usurf) > 0.000001 || fabs(flags.dPdx) > 0.000001)) {
  if (fabs(flags.Usurf) > 0.000001 || fabs(flags.dPdx) > 0.000001) {
    // initial a classical plane Couette-Poiseuille flow
    velocity_ini(u, flags.Usurf, flags.dPdx/nu);
    //u.addThePerturbations(((double)Nx+Nz)/6.0,0.001,spectralDecay);
  }
  else {
    // Perturb velocity field
    u.addThePerturbations(((double)Nx+Nz)/6.0,0.5,spectralDecay);
    u *= magnitude/L2Norm(u);
  }
  
  // Perturb scalar field
  //random1d(s, flags.kappa, 1);
  //random1d(s, 1);
  if (flags.iscalar > 0) {
    //if (flags.iscalar == 2) u.setToZero();
    scalar_ini(s, flags.BC, flags.bottomvalue, flags.topvalue);
    //s *= (0.25*magnitude)/L2Norm(s);
  }

// for dimension transfer only
/*
  Real mu, mp; 
  mu = 1.0*1.0e-6/(50.0*3.8e-5); // mu = delta_1*nu_2/(delta_2*nu_1)
  mp = mu*mu;
  u *= mu;
  q *= mp;
  u.b8Save("vini_revision.dat",t);
  q.b8Save("qini_revision.dat",t);
  exit(0);
*/
// -- end -- for dimension transfer only
  
  // Construct a variable time step object. See below for usage.
  TimeStep dt(dtini, dtmin, dtmax, dT, CFLmin, CFLmax);
  
  // Construct Navier-Stoke integrator, set integration method
  cout << "building DNS..." << flush;
  // Construct a DNS.
  DNS dns(u, U, nu, dt, flags, T0); // if use old way to add top surface velocity
  //DNS dns(u, nu, dt, flags, T0);

  cout << "done" << endl;

  // initialize the evolution file
  if (fexist("evolution.dat") == 0) u.evolution_ini();

  // load previous data if T0 is larger than zero
//  if (T0 > 0.0000001) {
//    if (fexist("vini.dat")) u.b8Read("vini.dat",&t);
//    if (fexist("qini.dat")) q.b8Read("qini.dat",&t);
//    if (flags.iscalar > 0 && fexist("sini.dat")) s.b8Read("sini.dat",&t);
//  }
  if (fexist("vini.dat")) {
     cout << "loading previous velocity field ..." << endl;
     u.b8Read("vini.dat",&t);
  }
  if (fexist("qini.dat")) q.b8Read("qini.dat",&t);
  if (flags.iscalar > 0 && fexist("sini.dat")) s.b8Read("sini.dat",&t);

  if (dt.adjust(dns.CFL())) {
     cout << "adjusting timestep" << endl;
     dns.reset_dt(dt);
  }
  
  // output geometry file in Plot3D format (for postprocessors including paraview)
  if (plot3d) u.plot3DXYZSave("geometry.xyz");

  // initialization for tecplot 10
  //u.tecplot10Open("velocity"); // one-file way

  time_t t1, t2;
  
  t1 = time(NULL);
  t2 = time(NULL);
  printf("t=%8.1f| CFL=%4.2f| K=%9.4e| Div(u)=%9.4e| CPS=%6.3f\n", 
    dns.time(), dns.CFL(), u.k(), divNorm(u), (t2-t1)/((double)dt.n()));

  Real Measures[62];
  Real Utop, UdL;
  Utop = fabs(flags.Usurf);
  if (flags.isurfp == 1) Utop +=fabs(flags.Up0)+fabs(flags.Up);
  UdL = Utop/(b-a);    
  
  int ftail, iout;
  iout = 0;
  assert(dT > 0.0 && T1 > T0);

  for (t=T0; t<T1; t += dT) {

    //cout << " t = " << t << " | CFL = " << dns.CFL() << " | K = " << u.k() << endl;
    //cout << " L2Norm(u) == " << L2Norm(u) << endl;
    //cout << "divNorm(u) == " << divNorm(u) << endl;
    //cout << "      dPdx == " << dns.dPdx() << endl;
    //cout << "     Ubulk == " << dns.Ubulk() << endl;
    //cout << "     error == " << u.error_vs_laminarEkman(U, flags.Usurf, nu, dns.Omega()) << endl;
    //cout << "     error == " << u.error_vs_laminarEkman(flags.Usurf, nu, dns.Omega()) << endl;
    //u.addThePerturbations(kf,0.5,spectralDecay);

    ftail = int(t+0.1);
    if (dT < 1.0) {
        ftail = iout;
        iout++;
    }
    
    // Save velocity,  pressure and density
    if (saveData) {
        u.b8Save("v."+i2sf(ftail),t);
        q.b8Save("q."+i2sf(ftail),t);
        if (flags.iscalar > 0) s.b8Save("s."+i2sf(ftail),t);
    }

    // output in Plot3D format (for postprocessors including paraview)
    //u.plot3DQSave("velocity"+i2sf(int(t))+".q",U,Reynolds,dns.time()); // if use old way to add top surface velocity
    //u.plot3DQSave("velocity"+i2sf(int(t))+".q",Reynolds,dns.time()); // if use new way to add top surface velocity
    //plot3DSave_2("velocity"+i2sf(int(t))+".q",s,u,U,Reynolds,dns.time()); // if use old way to add top surface velocity
    if (plot3d) plot3DSave_2("velocity"+i2sf(ftail)+".q",s,u,1.0/nu,dns.time()); // if use new way to add top surface velocity

    //u.tecplot10Save(dns.time()); // one-file way
    //u.tecplot10Save("velocity"+i2sf(int(t)),dns.time()); // multiple-file way
    if (tecplot) tecplot10Save("velocity"+i2sf(ftail),s,u,dns.time()); // multiple-file way

    // Output 2D spectrum
    u.saveSpectrum("spectrumxz."+i2sf(ftail), nu);
    //u.saveSpectrum("spectrumxz."+i2sf(ftail), nu, 5);

    // Output mean profiles using physical coordinates
    mean_profiles("mean."+i2sf(ftail),u,q,s,nu,flags.kappa,flags.g,flags.Omega,UdL,
          Measures,LRi,MProfile);

    //add_measures("add_measures."+i2sf(ftail),dns.time(),u,q,s,nu,flags.kappa,flags.g,flags.Omega);

    // Output U using wall coordinates
    //if (UPlus) u.saveU("U."+i2sf(ftail),nu); // for pressure driven flow
    //if (UPlus) u.saveUd("U."+i2sf(ftail),nu,Utop); // velocity defect
    if (UPlus) {
    
       if (flags.iscalar > 0) {
          saveUSd("U."+i2sf(ftail),MProfile,nu,flags.kappa);
       }
       else {
          saveUSd("U."+i2sf(ftail),MProfile,nu,0.0);
       }
       
    }

    // Output line sequences
    line_sequence("LS_u.dat", t, u, 0);
    line_sequence("LS_v.dat", t, u, 1);
    line_sequence("LS_w.dat", t, u, 2);
    line_sequence("LS_k.dat", t, u, 3);
    if (flags.iscalar > 0) {
    //line_sequence("LS_sp.dat", t, s, true);
    line_sequence("LS_s.dat", t, s);
    line_sequence_LRi("LS_lri.dat", t, LRi);
    }
    
    // Output evolutions
    u.evolution(dns.time(), dns.istep(), Measures);
    
    // Take n steps of length dt
    //for(int in=0; in<dt.n();in++) {
    //dns.advance(u, q, s, 1);
    //heat_error(dns.time(), dt.dt(), flags.kappa, s);
    //}
    //u.AscSave("u_ini.dat");
    if (flags.iscalar > 0) {
        dns.advance(u, q, s, dt.n());
        //dns.advance(u, q, s, 1);
     }
     else {
        dns.advance(u, q, dt.n());
        //dns.advance(u, q, 1);
     }
    //u.AscSave("u.dat");
    //exit(0);
   
    //heat_error(dns.time(), dt.dt(), flags.kappa, s);
    cout << endl;
    
    // If CFL number is outside the range CFLmin,CFLmax, adjust dt within 
    // limits dtmin,dtmax, and subject to dT/dt = n for some integer n.
    if (dt.adjust(dns.CFL())) {
        cout << "adjusting timestep" << endl;
        dns.reset_dt(dt);
    }

    t2 = time(NULL);
    
    printf("t=%8.1f| CFL=%4.2f| K=%9.4e| Div(u)=%9.4e| CPS=%6.3f\n", 
     dns.time(), dns.CFL(), u.k(), fabs(Measures[54]), (t2-t1)/((double)dt.n()));

    if (nRac_ > 1.0e-6) assert(Measures[55] < nRac_*657.511);
     
    t1 = t2;
    
  }

  ftail = int(t+0.1);
  if (dT < 1.0) {
  ftail = iout;
  iout++;
  }
  
  u.saveSpectrum("spectrumxz."+i2sf(ftail),nu);
  
  mean_profiles("mean."+i2sf(ftail),u,q,s,nu,flags.kappa,flags.g,flags.Omega,UdL,
        Measures,LRi,MProfile);
  
  line_sequence("LS_u.dat", t, u, 0);
  line_sequence("LS_v.dat", t, u, 1);
  line_sequence("LS_w.dat", t, u, 2);
  line_sequence("LS_k.dat", t, u, 3);
  if (flags.iscalar > 0) {
  line_sequence("LS_s.dat", t, s);
  line_sequence_LRi("LS_lri.dat", t, LRi);
  }

  u.evolution(dns.time(), dns.istep(), Measures);

  //if (UPlus) u.saveUd("U."+i2sf(ftail),nu,Utop);
  if (UPlus) {
    
     if (flags.iscalar > 0) {
        saveUSd("U."+i2sf(ftail),MProfile,nu,flags.kappa);
     }
     else {
        saveUSd("U."+i2sf(ftail),MProfile,nu,0.0);
     }
       
  }
  
  if (plot3d) plot3DSave_2("velocity"+i2sf(ftail)+".q",s,u,1.0/nu,dns.time()); // if use new way to add top surface velocity

  //u.tecplot10Save(dns.time()); // one-file way
  //u.tecplot10Close();  // one-file way
  //u.tecplot10Save("velocity"+i2sf(int(t)),t); // multiple-file way
  if (tecplot) tecplot10Save("velocity"+i2sf(ftail),s,u,t); // multiple-file way

  if (saveData) {
      u.b8Save("v."+i2sf(ftail),t);
      q.b8Save("q."+i2sf(ftail),t);
      if (flags.iscalar > 0) s.b8Save("s."+i2sf(ftail),t);
  }
  
}

