/* CRS (Channelflow with options of Rotation and Stratification), 
   revised Channelflow-1.1.2 (John F. Gibson) code including rotation and stratification
   Copyright (C) 2007-2008 Hao Lu

   Engine Research Center
   1500 Engineering Dr. #1011
   Madison, WI 53706

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <fstream>
#include <iomanip>
#include <string.h> // for strdupa
//#include <stddef.h> // for strtok#include 

#include "flowfield.h"
#include "symmetry.h"

using namespace std;

namespace channelflow {

FieldSymmetry::FieldSymmetry() 
  :
  sa_(1),
  sx_(1),
  sy_(1),
  sz_(1),
  fx_(0),
  fz_(0)
{}


FieldSymmetry::FieldSymmetry(int sa) 
  :
  sa_(sa),
  sx_(1),
  sy_(1),
  sz_(1),
  fx_(0.0),
  fz_(0.0)
{}

FieldSymmetry::FieldSymmetry(Real fx, Real fz) 
  :
  sa_(1),
  sx_(1),
  sy_(1),
  sz_(1),
  fx_(fx),
  fz_(fz)
{}

FieldSymmetry::FieldSymmetry(int sx, int sy, int sz, Real fx, Real fz, int sa)
  :
  sa_(sa),
  sx_(sx),
  sy_(sy),
  sz_(sz),
  fx_(fx),
  fz_(fz)
{
  if (abs(sa) != 1 || abs(sx) != 1 || abs(sy) != 1 || abs(sz) != 1 ) {
    cerr << "error in FieldSymmetry::FieldSymmetry(sa, sx, sy, sz, fx, fz) :" << endl;
    cerr << "sa,sx,sy,sz should be +/-1, but they are "
	 << sa << ',' << sx << ',' << sx << ',' << sz << endl;
    exit(1);
  }
}

FieldSymmetry::FieldSymmetry(bool sx, bool sy, bool sz, Real fx, Real fz, bool sa) 
  :
  sa_(sa ? -1:1),
  sx_(sx ? -1:1),
  sy_(sy ? -1:1),
  sz_(sz ? -1:1),
  fx_(fx),
  fz_(fz)
{}

int FieldSymmetry::s(int i) const {
  assert(i>=0 && i<3);
  int rtn = 0;
  switch (i) {
  case 0: rtn = sx_; break;
  case 1: rtn = sy_; break;
  case 2: rtn = sz_; break;
  default:
    cerr << "error in FieldSymmetry::s(int i) : i == " << i << " != 0,1, or 2" << endl;
    exit(1);
  }
  return rtn;
}

int FieldSymmetry::sign(int i) const {
  return s(i);
}

FlowField FieldSymmetry::operator()(const FlowField& u) const {
  FlowField v(u);
  v *= (*this);
  return v;
}


FieldSymmetry& FieldSymmetry::operator *= (const FieldSymmetry& s) {
  sa_ *= s.sa_;
  sx_ *= s.sx_;
  sy_ *= s.sy_;
  sz_ *= s.sz_;
  fx_ = fmod(fx_ + sx_*s.fx_, 1.0);
  fz_ = fmod(fx_ + sz_*s.fz_, 1.0);
  return *this;
}

bool operator == (const FieldSymmetry& p, const FieldSymmetry& q) {
  return (p.sa() == q.sa() &&
	  p.sx() == q.sx() &&
	  p.sy() == q.sy() && 
	  p.sz() == q.sz() && 
	  p.fx() == q.fx() && 
	  p.fz() == q.fz());
}

bool operator != (const FieldSymmetry& p, const FieldSymmetry& q) {
  return !(p == q);
}

FieldSymmetry operator* (const FieldSymmetry& p, const FieldSymmetry& q) {
  int sa = p.sa() * q.sa();
  int sx = p.sx() * q.sx();
  int sy = p.sy() * q.sy();
  int sz = p.sz() * q.sz();
  Real fx = fmod(p.fx() + p.sx()*q.fx(), 1.0);
  Real fz = fmod(p.fz() + p.sz()*q.fz(), 1.0);
  return FieldSymmetry(sx,sy,sz,fx,fz,sa);
}

FieldSymmetry inverse(const FieldSymmetry& s) {
  return FieldSymmetry(s.sx(), s.sy(), s.sz(), -s.fx(), -s.fz(), s.sa());
}

FlowField operator*(const FieldSymmetry& s, const FlowField& u) {
  return s(u);
}

ostream& operator<<(ostream& os, const FieldSymmetry& s) {
  streamsize p = os.precision();
  os << '(' 
     << setw(2) << setiosflags(ios::right) << s.sx() 
     << setw(3) << setiosflags(ios::right) << s.sy() 
     << setw(3) << setiosflags(ios::right) << s.sz() << "   "
     << resetiosflags(ios::adjustfield)
     << setw(p+6) << setiosflags(ios::left) << s.fx()
     << setw(p+6) << setiosflags(ios::left) << s.fz() 
     << setw(2) << setiosflags(ios::right) << s.sa() 
     << ')' << resetiosflags(ios::adjustfield);
  return os;
}

istream& operator>>(istream& is, FieldSymmetry& s) {
  char c;
  is >> c;
  if (c !='(' ) {
    cerr << "error in istream >> FieldSymmetry : expected '(' as opening character" << endl;
    cerr << "got '" << c << "' instead" << endl;
  }
  int sa,sx,sy,sz;
  Real fx,fz;
  is >> sx >> sy >> sz >> fx >> fz >> sa >> c;
  if (c !=')' )
    cferror("error in istream >> FieldSymmetry : expected ')' as closing character");
  
  s = FieldSymmetry(sx,sy,sz,fx,fz,sa);
  return is;
}

SymmetryList::SymmetryList() 
{}

SymmetryList::SymmetryList(int N) 
  :
  array<FieldSymmetry>(N)
{}

SymmetryList::SymmetryList(const string& filebase) {

  string filename(filebase);
  filename += string(".symms");
  ifstream is(filename.c_str(), ios::in);
  
  // If filebase.symms doesn't work, try filebase alone.
  if (!is.good()) 
    is.open(filebase.c_str(), ios::in);

  if (!is.good()) {
    cerr << "SymmetryList::SymmetryList(filebase) : can't open file " 
	 << filename << " or " << filebase << endl;
    exit(1);
  }
  int N;
  is >> N;
  resize(N);
  for (int n=0; n<N; ++n) 
    is >> (*this)[n];
}

void SymmetryList::save(const string& filebase) const {
  string filename = appendSuffix(filebase, ".symms");
  ofstream os(filename.c_str());
  
  int N = this->length();
  os << N << '\n';
  os << setprecision(17);
  for (int n=0; n<N; ++n) 
    os << (*this)[n] << '\n';
}


} // namespace channelflow
