/* CRS (Channelflow with options of Rotation and Stratification), 
   revised Channelflow-1.1.2 (John F. Gibson) code including rotation and stratification
   Copyright (C) 2007-2008 Hao Lu

   Engine Research Center
   1500 Engineering Dr. #1011
   Madison, WI 53706

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/* flowfield.h: Class for N-dim Fourier x Chebyshev x Fourier expansions
 * Channelflow-1.0
 *
 * Copyright (C) 2001-2007  John F. Gibson
 *
 * John F. Gibson
 * Center for Nonlinear Sciences
 * School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 *
 * gibson@cns.physics.gatech.edu
 * jfg@member.fsf.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */

#ifndef CHANNELFLOW_SYMMETRY_H
#define CHANNELFLOW_SYMMETRY_H

#include <fftw3.h>
#include "mathdefs.h"
#include "vector.h"
#include "chebyshev.h"
#include "basisfunc.h"
#include "realprofile.h"

namespace channelflow {

// FieldSymmetry s(sx,sy,sz,fx,fz,sa) takes fx,fz in [0,1] and sa,sx,sy,sz in {-1,1}
// s : (u,v,w)(x,y,z) -> (sa sx u, sa sy v, sa sz w)(sx x + fx Lx, sy y, sz z + fz Lz)

class FlowField;

class FieldSymmetry {
public:
  FieldSymmetry();
  //FieldSymmetry(std::istream& is);
  //FieldSymmetry(const std::string& filebase);
  FieldSymmetry(int sa);                 // sign change in (u,v,w), sa = +/-1
  FieldSymmetry(Real fx, Real fz);       // pure translate
  FieldSymmetry(int  sx, int  sy, int  sz, Real fx=0.0, Real fz=0.0, int sa=1);
  FieldSymmetry(bool sx, bool sy, bool sz, Real fx=0.0, Real fz=0.0, bool sa=false);

  FlowField operator()(const FlowField& u) const;
  FieldSymmetry& operator *=(const FieldSymmetry& s);

  inline int sa() const;
  inline int sx() const;
  inline int sy() const;
  inline int sz() const;
  inline Real fx() const;
  inline Real fz() const;
  int s(int i) const;
  int sign(int i) const;

  void save(const std::string& filebase) const;
  //Real error(const FlowField& u) const; // L2Dist(u, (this)(u))
  //FieldSymmetry optimalPhase(const FlowField& u) const; 

private:
  int sa_;   // (u,v,w)(x,y,z) -> (sa u, sa v, sa w)(x,y,z)
  int sx_;   // (u,v,w)(x,y,z) -> (sx u, v,w)(sx x, y,z) 
  int sy_;   // (u,v,w)(x,y,z) -> (u, sy v,w)(x, sy y,z) 
  int sz_;   // (u,v,w)(x,y,z) -> (u,v, sz w)(x,y, sz z) 
  Real fx_;  // (u,v,w)(x,y,z) -> (u,v,w)(x+fx*Lx,y,z)
  Real fz_;  // (u,v,w)(x,y,z) -> (u,v,w)(x,y,z+fz/Lz)
};

bool operator ==(const FieldSymmetry& p, const FieldSymmetry& q);
bool operator !=(const FieldSymmetry& p, const FieldSymmetry& q);

std::istream& operator>>(std::istream& is, FieldSymmetry& s);
std::ostream& operator<<(std::ostream& os, const FieldSymmetry& s);

//FieldSymmetry optimizePhase(const FlowField& u, const FieldSymmetry& s);
FieldSymmetry inverse(const FieldSymmetry& s);
FlowField operator*(const FieldSymmetry& s, const FlowField& u);// same as s(u)

class SymmetryList : public array<FieldSymmetry> {
 public:
  SymmetryList();
  SymmetryList(int n);
  SymmetryList(const std::string& filebase);
  
  void save(const std::string& filebase) const;
};

inline Real FieldSymmetry::fx() const {return fx_;}
inline Real FieldSymmetry::fz() const {return fz_;}
inline int  FieldSymmetry::sa() const {return sa_;}
inline int  FieldSymmetry::sx() const {return sx_;}
inline int  FieldSymmetry::sy() const {return sy_;}
inline int  FieldSymmetry::sz() const {return sz_;}

FieldSymmetry operator* (const FieldSymmetry& s1, const FieldSymmetry& s2);

}

#endif 
