/* CRS (Channelflow with options of Rotation and Stratification), 
   revised Channelflow-1.1.2 (John F. Gibson) code including rotation and stratification
   Copyright (C) 2007-2008 Hao Lu

   Engine Research Center
   1500 Engineering Dr. #1011
   Madison, WI 53706

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iomanip>
#include "tausolver_bc.h"

using namespace std;

namespace channelflow {

const Real MINIMUM_DISCRIMINANT = 1e-4;
const Real EPSILON = 1e-7;

TauSolver_BC::TauSolver_BC() 
  :
  N_(0),
  Nb_(0),
  kx_(0),
  kz_(0),
  two_pi_kxLx_(0),
  two_pi_kzLz_(0),
  kappa2_(0),
  a_(0),
  b_(0),
  lambda_(0),
  nu_(0),
  A_Helmholtz_(),
  B_Helmholtz_(),
  DsB_(0)
{}

TauSolver_BC::TauSolver_BC(BoundaryC BC, int kx, int kz, Real Lx, Real Lz, 
	Real a, Real b, Real lambda, Real nu, int nChebyModes)
  :
  BC_(BC),
  N_(nChebyModes),
  Nb_(nChebyModes-1),
  kx_(kx),
  kz_(kz),
  two_pi_kxLx_(2*pi*kx/Lx),
  two_pi_kzLz_(2*pi*kz/Lz),
  kappa2_(4*square(pi)*(square(kx/Lx)+square(kz/Lz))),
  a_(a),
  b_(b),
  lambda_(lambda),
  nu_(nu),
  DsB_(0)
{

if (BC_ == DbDt) A_Helmholtz_ = HelmholtzSolver(N_, a_, b_, lambda_, nu_);

if (BC_ == NbNt) A_Helmholtz_ = HelmholtzSolver(true, N_, a_, b_, lambda_, nu_);

if (BC_ == DbNt) 
	{
	A_Helmholtz_ = HelmholtzSolver(N_, a_, b_, lambda_, nu_);
	B_Helmholtz_ = HelmholtzSolver(N_, a_, b_, lambda_, nu_);

	sB_ = ChebyCoeff(N_, a_, b_, Spectral);

	ChebyCoeff zero(N_, a_, b_, Spectral);
	zero.setToZero();
	B_Helmholtz_.solve(sB_, zero, 0.0, 1.0);

	ChebyCoeff dBdy(N_, a_, b_, Spectral);
	diff(sB_, dBdy);
	DsB_ = dBdy.eval_b();

    assert((abs(DsB_) > MINIMUM_DISCRIMINANT));
	}

if (BC_ == NbDt) 
	{
	A_Helmholtz_ = HelmholtzSolver(N_, a_, b_, lambda_, nu_);
	B_Helmholtz_ = HelmholtzSolver(N_, a_, b_, lambda_, nu_);

	sB_ = ChebyCoeff(N_, a_, b_, Spectral);

	ChebyCoeff zero(N_, a_, b_, Spectral);
	zero.setToZero();
	B_Helmholtz_.solve(sB_, zero, 1.0, 0.0);

	ChebyCoeff dBdy(N_, a_, b_, Spectral);
	diff(sB_, dBdy);
	DsB_ = dBdy.eval_a();

    assert((abs(DsB_) > MINIMUM_DISCRIMINANT));
	}

}
  
void TauSolver_BC::solve(ChebyCoeff& s, const ChebyCoeff& Rs, Real sa, Real sb) const
{

if (BC_ == DbDt)  A_Helmholtz_.solve(s, Rs, sa, sb);

if (BC_ == NbNt)  A_Helmholtz_.solve_Neumann(s, Rs, sa, sb);

if (BC_ == DbNt) 
	{
	A_Helmholtz_.solve(s, Rs, sa, 0.0);

	ChebyCoeff dAdy(N_, a_, b_, Spectral);
	diff(s, dAdy);
	
	Real delta = (sb-dAdy.eval_b())/DsB_;

	for (int i=0; i<N_; ++i) s[i] += delta*sB_[i];
	}

if (BC_ == NbDt) 
	{
	A_Helmholtz_.solve(s, Rs, 0.0, sb);

	ChebyCoeff dAdy(N_, a_, b_, Spectral);
	diff(s, dAdy);
	
	Real delta = (sa-dAdy.eval_a())/DsB_;

	for (int i=0; i<N_; ++i) s[i] += delta*sB_[i];
	}
	
}

} //namespace channelflow

