/* CRS (Channelflow with options of Rotation and Stratification), 
   revised Channelflow-1.1.2 (John F. Gibson) code including rotation and stratification
   Copyright (C) 2007-2008 Hao Lu

   Engine Research Center
   1500 Engineering Dr. #1011
   Madison, WI 53706

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
/* flowfield.h: Class for N-dim Fourier x Chebyshev x Fourier expansions
 * Channelflow-1.0
 *
 * Copyright (C) 2001-2007  John F. Gibson
 *
 * John F. Gibson
 * Center for Nonlinear Sciences
 * School of Physics
 * Georgia Institute of Technology
 * Atlanta, GA 30332-0430
 *
 * gibson@cns.physics.gatech.edu
 * jfg@member.fsf.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, U
 */

/* modified by Hao Lu (haolu@wisc.edu) on June 18th, 2007
 * for using it to study rotating turbulence
 */

#include <fstream>
#include <iomanip>
#include <string.h> // for strdupa
#include <stddef.h> // for strtok
#include <stdlib.h>

#include "flowfield.h"
#include "diffops.h"
#include "symmetry.h"
#include "addition.h"
#include "TECXXX.h"

using namespace std;

namespace channelflow {

const Real EPSILON=1e-4;

FlowField::FlowField()
  :
  Nx_(0),
  Ny_(0),
  Nz_(0),
  Nzpad_(0),
  Nzpad2_(0),
  Nd_(0),
  Lx_(0),
  Lz_(0),
  a_(0),
  b_(0),
  dealiased_(false),
  rdata_(0),
  cdata_(0),
  scratch_(0),
  xzstate_(Spectral),
  ystate_(Spectral),
  xz_plan_(0),
  xz_iplan_(0),
  y_plan_(0),
  iles_(0),
  tau_(0),
  k_(0),
  Nmax_(0)
{}

FlowField::FlowField(int Nx, int Ny, int Nz, int Nd, Real Lx, Real Lz, 
             Real a, Real b, fieldstate xzstate, fieldstate ystate, 
             uint fftw_flags) 
  :
  Nx_(0),
  Ny_(0),
  Nz_(0),
  Nzpad_(0),
  Nzpad2_(0),
  Nd_(0),
  Lx_(0),
  Lz_(0),
  a_(0),
  b_(0),
  dealiased_(false),
  rdata_(0),
  cdata_(0),
  scratch_(0),
  xzstate_(xzstate),
  ystate_(ystate),
  xz_plan_(0),
  xz_iplan_(0),
  y_plan_(0),
  iles_(0),
  tau_(0),
  k_(0),
  Nmax_(0)
{
  // This isn't in classic C++ initialization style, but it consolidates
  // all resize and initialization code, and the relative overhead is neglgble.
  resize(Nx, Ny, Nz, Nd, Lx, Lz, a, b, fftw_flags);
}

FlowField::FlowField(int Nx, int Ny, int Nz, int Nd, int tensorOrder, Real Lx, 
             Real Lz, Real a, Real b, fieldstate xzstate, 
             fieldstate ystate, uint fftw_flags) 
  :
  Nx_(0),
  Ny_(0),
  Nz_(0),
  Nzpad_(0),
  Nzpad2_(0),
  Nd_(pow(Nd,tensorOrder)),
  Lx_(0),
  Lz_(0),
  a_(0),
  b_(0),
  dealiased_(false),
  rdata_(0),
  cdata_(0),
  scratch_(0),
  xzstate_(xzstate),
  ystate_(ystate),
  xz_plan_(0),
  xz_iplan_(0),
  y_plan_(0),
  iles_(0),
  tau_(0),
  k_(0),
  Nmax_(0)
{
  // This isn't in classic C++ initialization style, but it consolidates
  // all resize and initialization code, and the relative overhead is neglgble.
  resize(Nx, Ny, Nz, Nd_, Lx, Lz, a, b, fftw_flags);
}

FlowField::FlowField(const FlowField& f) 
  :
  Nx_(0),
  Ny_(0),
  Nz_(0),
  Nzpad_(0),
  Nzpad2_(0),
  Nd_(0),
  Lx_(0),
  Lz_(0),
  a_(0),
  b_(0),
  dealiased_(f.dealiased_),
  rdata_(0),
  cdata_(0),
  scratch_(0),
  xzstate_(Spectral),
  ystate_(Spectral),
  xz_plan_(0),
  xz_iplan_(0),
  y_plan_(0),
  iles_(0),
  tau_(0),
  k_(0),
  Nmax_(0)
{
  resize(f.Nx_, f.Ny_, f.Nz_, f.Nd_, f.Lx_, f.Lz_, f.a_, f.b_);

  setState(f.xzstate_, f.ystate_);
  dealiased_ = f.dealiased_;
  int N = Nd_*Ny_*Nx_*Nzpad_;
  for (int i=0; i<N; ++i)
    rdata_[i] = f.rdata_[i];

}

FlowField::FlowField(const string& filebase) 
  :
  Nx_(0),
  Ny_(0),
  Nz_(0),
  Nzpad_(0),
  Nzpad2_(0),
  Nd_(0),
  Lx_(0.0),
  Lz_(0.0),
  a_(0.0),
  b_(0.0),
  dealiased_(false),
  rdata_(0),
  cdata_(0),
  scratch_(0),
  xzstate_(Spectral),
  ystate_(Spectral),
  xz_plan_(0),
  xz_iplan_(0),
  y_plan_(0),
  iles_(0),
  tau_(0),
  k_(0),
  Nmax_(0)
{
  
  string filename(filebase);
  filename += string(".ff");
  ifstream is(filename.c_str(), ios::in | ios::binary);

  // If filebase.ff doesn't work, try filebase alone.
  if (!is.good()) {
    is.close();
    is.open(filebase.c_str(), ios::in | ios::binary);
    if (is.good()) 
      filename = filebase;
    else {
      cerr << "FlowField::FlowField(filebase) : can't open file " 
       << filename << " or " << filebase << endl;
      exit(1);
    }
  }

  int major_version, minor_version, update_version;
  read(is, major_version);
  read(is, minor_version);
  read(is, update_version);
  /*********************8
  if (major_version < 1 && minor_version <= 9 && update_version < 16) {
    cerr << "FlowField::FlowField(filebase) : file " << filename
     << " is not in version > 0.9.16 format.\n"
     << "Version appears to be " << major_version << '.' 
     << minor_version << '.' << update_version << endl;
    cerr << "Try using FlowField::FlowField(filebase, chflow_version) constructor.\n";
    exit(1);
  }
  **********************/
  read(is, Nx_);
  read(is, Ny_);
  read(is, Nz_);
  read(is, Nd_);
  read(is, xzstate_);
  read(is, ystate_);
  read(is, Lx_);
  read(is, Lz_);
  read(is, a_);
  read(is, b_);
  read(is, dealiased_);
  resize(Nx_, Ny_, Nz_, Nd_, Lx_, Lz_, a_, b_);

  // Read data only for non-aliased modes, assume 0 for aliased.
  if (dealiased_ == true && xzstate_ == Spectral) {
    int Nxd=2*(Nx_/6);
    int Nzd=2*(Nz_/3)+1;

    // In innermost loop, array index is (nz + Nzpad2_*(nx + Nx_*(ny + Ny_*i))),
    // which is the same as the FlowField::flatten function.
    for (int i=0; i<Nd_; ++i) {
      for (int ny=0; ny<Ny_; ++ny) {

    for (int nx=0; nx<=Nxd; ++nx) {
      for (int nz=0; nz<=Nzd; ++nz)
        read(is, rdata_[flatten(nx,ny,nz,i)]);
      for (int nz=Nzd+1; nz<Nzpad_; ++nz)
        rdata_[flatten(nx,ny,nz,i)] = 0.0;
    }
    for (int nx=Nxd+1; nx<=Nxd; ++nx) 
      for (int nz=0; nz<=Nzpad_; ++nz)
        rdata_[flatten(nx,ny,nz,i)] = 0.0;
    
    for (int nx=Nx_-Nxd; nx<Nx_; ++nx) {
      for (int nz=0; nz<=Nzd; ++nz)
        read(is, rdata_[flatten(nx,ny,nz,i)]);
      for (int nz=Nzd+1; nz<Nzpad_; ++nz)
        rdata_[flatten(nx,ny,nz,i)] = 0.0;
    }
      }
    }
  }
  else {
    int N = Nd_*Ny_*Nx_*Nzpad_;
    for (int i=0; i<N; ++i)
      read(is, rdata_[i]);
  }
}

FlowField::FlowField(const string& filebase, int major, int minor, int update)
  :
  Nx_(0),
  Ny_(0),
  Nz_(0),
  Nzpad_(0),
  Nzpad2_(0),
  Nd_(0),
  Lx_(0.0),
  Lz_(0.0),
  a_(0.0),
  b_(0.0),
  dealiased_(false),
  rdata_(0),
  cdata_(0),
  scratch_(0),
  xzstate_(Spectral),
  ystate_(Spectral),
  xz_plan_(0),
  xz_iplan_(0),
  y_plan_(0),
  iles_(0),
  tau_(0),
  k_(0),
  Nmax_(0)
{
  
  string filename(filebase);
  filename += string(".ff");
  ifstream is(filename.c_str(), ios::in | ios::binary);
  
  // If filebase.ff doesn't work, try filebase alone.
  if (!is.good()) 
    is.open(filebase.c_str(), ios::in | ios::binary);

  if (!is.good()) {
    cerr << "FlowField::FlowField(filebase, major,minor,update) : can't open file " 
     << filename << " or " << filebase << endl;
    exit(1);
  }

  if (major > 0 || minor > 9) {
    cerr << "FlowField::FlowField(filebase, major,minor,update) :\n"
     << major << '.' << minor << '.' << update << " is higher than 0.9.x" << endl;
    exit(1);
  }
  
  if (update >= 16) {
    (*this) = FlowField(filebase); // not efficient but saves code redundancy
    return;
  }
  else {
    is.read((char*) &Nx_, sizeof(int));
    is.read((char*) &Ny_, sizeof(int));
    is.read((char*) &Nz_, sizeof(int));
    is.read((char*) &Nd_, sizeof(int));
    is >> xzstate_;
    is >> ystate_;
    is.read((char*) &Lx_, sizeof(Real));
    is.read((char*) &Lz_, sizeof(Real));
    is.read((char*) &a_, sizeof(Real));
    is.read((char*) &b_, sizeof(Real));
    char s;
    is.get(s);
    dealiased_ = (s=='1') ? true : false;

    resize(Nx_, Ny_, Nz_, Nd_, Lx_, Lz_, a_, b_);

    // Read data only for non-aliased modes, assume 0 for aliased.
    if (dealiased_ == true && xzstate_ == Spectral) {
      int Nxd=2*(Nx_/6);
      int Nzd=2*(Nz_/3)+1;

      // In innermost loop, array index is (nz + Nzpad2_*(nx + Nx_*(ny+Ny_*i)))
      // which is the same as the FlowField::flatten function.
      for (int i=0; i<Nd_; ++i) {
    for (int ny=0; ny<Ny_; ++ny) {

      for (int nx=0; nx<=Nxd; ++nx) {
        for (int nz=0; nz<=Nzd; ++nz)
          is.read((char*) (rdata_ + flatten(nx,ny,nz,i)), sizeof(Real));
        for (int nz=Nzd+1; nz<Nzpad_; ++nz)
          rdata_[flatten(nx,ny,nz,i)] = 0.0;
      }
      for (int nx=Nxd+1; nx<=Nxd; ++nx) 
        for (int nz=0; nz<=Nzpad_; ++nz)
          rdata_[flatten(nx,ny,nz,i)] = 0.0;
    
      for (int nx=Nx_-Nxd; nx<Nx_; ++nx) {
        for (int nz=0; nz<=Nzd; ++nz)
          is.read((char*) (rdata_ + flatten(nx,ny,nz,i)), sizeof(Real));
        for (int nz=Nzd+1; nz<Nzpad_; ++nz)
          rdata_[flatten(nx,ny,nz,i)] = 0.0;
      }
    }
      }
    }
    else {
      int N = Nd_*Ny_*Nx_*Nzpad_;
      for (int i=0; i<N; ++i)
    is.read((char*) (rdata_ + i), sizeof(Real));
    }
  }
}

Vector FlowField::xgridpts() const {
  Vector xpts(Nx_);
  for (int nx=0; nx<Nx_; ++nx)
    xpts[nx] = x(nx);
  return xpts;
}
Vector FlowField::ygridpts() const {
  Vector ypts(Ny_);
  Real c = 0.5*(b_+a_);
  Real r = 0.5*(b_-a_);
  Real piN = pi/(Ny_-1);
  for (int ny=0; ny<Ny_; ++ny)
    ypts[ny] = c + r*cos(piN*ny);
  return ypts;
}
Vector FlowField::zgridpts() const {
  Vector zpts(Nz_);
  for (int nz=0; nz<Nz_; ++nz)
    zpts[nz] = z(nz);
  return zpts;
}

FlowField& FlowField::operator=(const FlowField& f) {
  resize(f.Nx_, f.Ny_, f.Nz_, f.Nd_, f.Lx_, f.Lz_, f.a_, f.b_);
  setState(f.xzstate_, f.ystate_);
  dealiased_ = f.dealiased_;
  int Ntotal = Nx_ * Ny_ * Nzpad_ * Nd_;
  for (int i=0; i<Ntotal; ++i)
    rdata_[i] = f.rdata_[i];
  return *this;
}

FlowField::~FlowField() {
  if (rdata_)    fftw_free(rdata_);
  if (scratch_)  fftw_free(scratch_);
  if (y_plan_)   fftw_destroy_plan(y_plan_);
  if (xz_plan_)  fftw_destroy_plan(xz_plan_);
  if (xz_iplan_) fftw_destroy_plan(xz_iplan_);
}

bool FlowField::isNull() {
  return (Nx_ == 0 && Ny_ == 0 && Nz_ == 0 && Nd_ == 0);
}

void FlowField::reconfig(const FlowField& u, uint fftw_flags) {
  resize(u.Nx(), u.Ny(), u.Nz(), u.Nd(), u.Lx(),u.Lz(),u.a(),u.b(),fftw_flags);
}

void FlowField::resize(int Nx, int Ny, int Nz, int Nd, Real Lx, Real Lz, 
               Real a, Real b, uint fftw_flags) {

  assert(Nx>=0);
  assert(Ny>=0);
  assert(Nz>=0);
  assert(Nd>=0);
  assert(Lx>=0);
  assert(Lz>=0);
  assert(b >= a);
   
  Nx_ = Nx;
  Ny_ = Ny;
  Nz_ = Nz;
  Nd_ = Nd;
  Lx_ = Lx;
  Lz_ = Lz;
  a_ = a;
  b_ = b;

  Nzpad_ = 2*(Nz_/2+1);
  Nzpad2_ = Nz_/2+1;

  // INEFFICIENT if geometry doesn't change. Should check.
  if (rdata_) fftw_free(rdata_);
  rdata_ = (Real*) fftw_malloc((Nx_*Ny_*Nzpad_*Nd_)*sizeof(Real));
  cdata_ = (Complex*) rdata_;

  if (scratch_) fftw_free(scratch_);
  scratch_ = (Real*) fftw_malloc(Ny_*sizeof(Real));

  fftw_initialize(fftw_flags);

  int N = Nd_*Ny_*Nx_*Nzpad_;
  for (int i=0; i<N; ++i)  rdata_[i] = 0.0;
}
  
void FlowField::fftw_initialize(uint fftw_flags) {
  fftw_flags = fftw_flags | FFTW_DESTROY_INPUT;
  fftw_flags = FFTW_MEASURE | FFTW_DESTROY_INPUT;

  if (y_plan_)   fftw_destroy_plan(y_plan_);
  if (xz_plan_)  fftw_destroy_plan(xz_plan_);
  if (xz_iplan_) fftw_destroy_plan(xz_iplan_);

  // Initialize xz plan
  if (Nx_ !=0 && Nz_ !=0) {
    fftw_complex* fcdata = (fftw_complex*) cdata_;

    const int howmany = Ny_*Nd_;
    const int rank = 2;

    // These params describe the structure of the real-valued array 
    int real_n[rank];
    real_n[0] = Nx_;
    real_n[1] = Nz_;
    int real_embed[rank];
    real_embed[0] = Nx_;
    real_embed[1] = Nzpad_;
    const int real_stride = 1;
    const int real_dist = Nx_*Nzpad_;

    // These params describe the structure of the complex-valued array 
    int cplx_n[rank];
    cplx_n[0] = Nx_;
    cplx_n[1] = Nzpad2_;
    int cplx_embed[rank];
    cplx_embed[0] = Nx_;
    cplx_embed[1] = Nzpad2_;
    const int cplx_stride = 1;
    const int cplx_dist = Nx_*Nzpad2_;

    // Real -> Complex transform parameters
    xz_plan_ = 
      fftw_plan_many_dft_r2c(rank, real_n, howmany, 
                 rdata_, real_embed, real_stride, real_dist, 
                 fcdata, cplx_embed, cplx_stride, cplx_dist, 
                 fftw_flags);

    xz_iplan_ = 
      fftw_plan_many_dft_c2r(rank, real_n, howmany, 
                 fcdata, cplx_embed, cplx_stride, cplx_dist, 
                 rdata_, real_embed, real_stride, real_dist, 
                 fftw_flags);
  }
  else {
    xz_plan_ = 0;
    xz_iplan_ = 0;
  }

  // Initialize y plan
  if (Ny_ >= 2) 
    y_plan_  = fftw_plan_r2r_1d(Ny_,scratch_,scratch_,FFTW_REDFT00,fftw_flags);
  else 
    y_plan_  = 0;

}

void FlowField::rescale(Real Lx, Real Lz) {
  assertState(Spectral, Spectral);
  Real scalev = Lx_/Lx;
  Real scalew = (Lx*Lz_)/(Lz*Lx_);
  
  for (int ny=0; ny<numYmodes(); ++ ny)
    for (int mx=0; mx<Mx(); ++mx)
      for (int mz=0; mz<Mz(); ++ mz) {
    cmplx(mx,ny,mz,1) *= scalev;
    cmplx(mx,ny,mz,2) *= scalew;
      }
  Lx_ = Lx;
  Lz_ = Lz;
}

void FlowField::interpolate(const FlowField& ff) {
  FlowField& g = *this;
  FlowField& f = (FlowField&) ff;

  fieldstate fxstate = f.xzstate();
  fieldstate fystate = f.ystate();
  fieldstate gxstate = g.xzstate();
  fieldstate gystate = g.ystate();
  
  // Always check these, debugging or not. 
  if (f.Lx() != g.Lx()  || f.Lz() != g.Lz() || f.a() != g.a()  ||
      f.b() != g.b() || f.Nd() != g.Nd()) {
    cerr << "FlowField::interpolate(const FlowField& f) error:\n"
     << "FlowField doesn't match argument f geometrically.\n";
    exit(1);
  }

  const int fNy = f.Ny();
  const int gNy = g.Ny();
  const int Nd = f.Nd();
  g.setToZero();

  ComplexChebyCoeff fprof(f.Ny(), a_, b_, Spectral);
  ComplexChebyCoeff gprof(g.Ny(), a_, b_, Spectral);

  const int kxmax = lesser(f.kxmax(), g.kxmax());
  const int kzmax = lesser(f.kzmax(), g.kzmax());
  const int kxmin = Greater(f.kxmin(), g.kxmin());

  f.makeSpectral();
  g.setState(Spectral, Spectral);
  g.setToZero();

  for (int i=0; i<Nd; ++i) {

    // If going from smaller Chebyshev expansion to larger
    // Just copy the spectral coefficients and loop in fastest order
    if (fNy <= gNy) {
      for (int ny=0; ny<fNy; ++ny) 
    for (int kx=kxmin; kx<kxmax; ++kx) {
      int fmx = f.mx(kx);
      int gmx = g.mx(kx);
      for (int kz=0; kz<=kzmax; ++kz) {
        int fmz = f.mz(kz);
        int gmz = g.mz(kz);
        g.cmplx(gmx,ny,gmz,i) = f.cmplx(fmx,ny,fmz,i);
    }
      }
    }
    // Going from larger Chebyshev expansion to smaller
    // Interpolate the spectral coefficients. That requires an inner loop in y.
    else {
      for (int kx=kxmin; kx<kxmax; ++kx) {
    int fmx = f.mx(kx);
    int gmx = g.mx(kx);
    for (int kz=0; kz<=kzmax; ++kz) {
      int fmz = f.mz(kz);
      int gmz = g.mz(kz);
      for (int ny=0; ny<fNy; ++ny) 
        fprof.set(ny, f.cmplx(fmx,ny,fmz,i));
      gprof.interpolate(fprof);
      for (int ny=0; ny<gNy; ++ny) 
        g.cmplx(gmx,ny,gmz,i) = gprof[ny];
    }        
      }
    }

    // Fourier discretizations go from -Mx/2+1 <= kx <= Mx/2
    // So the last mode kx=Mx/2 has an implicit -kx counterpart that is
    // not included in the Fourier representation. For f, this mode is
    // implicitly/automatically included in the Fourier transforms. 
    // But if g has more x modes than f, we need to assign
    // g(-kxmax) = conj(f(kxmax)) explicitly.
    if (g.Nx() > f.Nx()) {
      int kx = f.kxmax();
      int fmx = f.mx(kx);
      int gmx = g.mx(-kx);

      if (fNy <= gNy) {
    for (int ny=0; ny<fNy; ++ny) {
      for (int kz=f.kzmin(); kz<=f.kzmax(); ++kz) {
        int fmz = f.mz(kz);
        int gmz = g.mz(kz);
        g.cmplx(gmx,ny,gmz,i) = f.cmplx(fmx,ny,fmz,i);
      }
    }
      }
      else {
    for (int kz=f.kzmin(); kz<=f.kzmax(); ++kz) {
      int fmz = f.mz(kz);
      int gmz = g.mz(kz);
      for (int ny=0; ny<fNy; ++ny) 
        fprof.set(ny, f.cmplx(fmx,ny,fmz,i));
      gprof.interpolate(fprof);
      for (int ny=0; ny<gNy; ++ny) 
        g.cmplx(gmx,ny,gmz,i) = gprof[ny];
    }        
      }
    }
  }
    
  f.makeState(fxstate, fystate);
  g.makeState(gxstate, gystate);
}


void FlowField::optimizeFFTW(uint fftw_flags) {
  fftw_flags = fftw_flags | FFTW_DESTROY_INPUT;
  fftw_initialize(fftw_flags);
}

ComplexChebyCoeff FlowField::profile(int mx, int mz, int i) const {
  ComplexChebyCoeff rtn(Ny_, a_, b_, ystate_);
  if (xzstate_== Spectral) {
    for (int ny=0; ny<Ny_; ++ny)
      rtn.set(ny, cmplx(mx, ny, mz, i));
  }
  else 
    for (int ny=0; ny<Ny_; ++ny)
      rtn.re[ny] = (*this)(mx, ny, mz, i);
    
  return rtn;
}

BasisFunc FlowField::profile(int mx, int mz) const {
  assert(xzstate_== Spectral);
  BasisFunc rtn(Nd_, Ny_, kx(mx), kz(mz), Lx_, Lz_, a_, b_, ystate_);
  for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny)
      rtn[i].set(ny, cmplx(mx, ny, mz, i));
  return rtn;
}

// (u,v,w)(x,y,z) -> (u,v,-w)(x, y, -z)
FlowField& FlowField::reflect() {
  assert(Nd_ == 3);
  fieldstate xzs = xzstate_;
  makeState(Spectral, ystate_);

  // Looping is nonobvious. These bounds on kx avoid the kxmax cosine mode,
  // which is either not present (Nx even), or doesn't have a symmetric 
  // counterpart (Nx odd). 2007-02-23 notes.
  Complex tmp;
  for (int i=0; i<Nd_; ++i) {
    Real isign = (i==2) ? -1.0 : 1.0;
    for (int ny=0; ny<Ny_; ++ny) {
      
      for (int kx_m = kxmin(); kx_m<=0; ++kx_m) {
    int mx_m = mx(kx_m);  // mx minus
    int mx_p = mx(-kx_m); // mx plus

    for (int mz=0; mz<Mz(); ++mz) {
      tmp = cmplx(mx_m,ny,mz,i);
      cmplx(mx_m,ny,mz,i) = isign*conj(cmplx(mx_p,ny,mz,i));
      cmplx(mx_p,ny,mz,i) = isign*conj(tmp);
    }
      }
    }
  }
  makeState(xzs, ystate_);
  return *this;
}

// (u,v,w)(x,y,z) -> (-u,-v,w)(-x,-y, z)
FlowField& FlowField::rotate() {
  assert(Nd_ == 3);
  fieldstate xzs = xzstate_;
  fieldstate ys  = ystate_;
  makeState(Spectral, Spectral);

  // Looping is nonobvious. These bounds on kx avoid the kxmax cosine mode,
  // which is either not present (Nx even), or doesn't have a symmetric 
  // counterpart (Nx odd). 2007-02-23 notes.
  Complex tmp;
  for (int i=0; i<Nd_; ++i) {
    int isign = (i==2) ? 1 : -1;

    for (int ky=0; ky<Ny_; ++ky) {
      int ysign = (ky % 2 == 0) ? 1 : -1;  

      for (int kx_m = kxmin(); kx_m<=0; ++kx_m) {
    int mx_m = mx(kx_m);  // mx minus
    int mx_p = mx(-kx_m); // mx plus
    int xsign = (kx_m % 2 == 0) ? 1 : -1;
    Real sig = Real(isign*ysign*xsign);

    for (int mz=0; mz<Mz(); ++mz) {
      tmp = cmplx(mx_m,ky,mz,i);
      cmplx(mx_m,ky,mz,i) = sig*cmplx(mx_p,ky,mz,i);
      cmplx(mx_p,ky,mz,i) = sig*tmp;
    }
      }
    }
  }
  makeState(xzs, ys);
  return *this;
}

// (u,v,w)(x,y,z) -> (u,v,-w)(x+Lx/2, y, -z)
FlowField& FlowField::shiftReflect() {
  assert(Nd_ == 3);
  fieldstate xzs = xzstate_;
  makeState(Spectral, ystate_);

  // Looping is nonobvious. These bounds on kx avoid the kxmax cosine mode,
  // which is either not present (Nx even), or doesn't have a symmetric 
  // counterpart (Nx odd). 2007-02-23 notes.
  Complex tmp;
  for (int i=0; i<Nd_; ++i) {
    int isign = (i==2) ? -1 : 1;
    for (int ny=0; ny<Ny_; ++ny) {
      
      for (int kx_m = kxmin(); kx_m<=0; ++kx_m) {
    int mx_m = mx(kx_m);  // mx minus
    int mx_p = mx(-kx_m); // mx plus
    int xsign = (kx_m % 2 == 0) ? 1 : -1;
    Real sig = Real(isign * xsign);

    for (int mz=0; mz<Mz(); ++mz) {
      tmp = cmplx(mx_m,ny,mz,i);
      cmplx(mx_m,ny,mz,i) = sig*conj(cmplx(mx_p,ny,mz,i));
      cmplx(mx_p,ny,mz,i) = sig*conj(tmp);
    }
      }
    }
  }
  makeState(xzs, ystate_);
  return *this;
}

// (u,v,w)(x,y,z) -> (-u,-v,w)(Lx/2-x, -y, Lz/2+z)
FlowField& FlowField::shiftRotate() {
  assert(Nd_ == 3);
  fieldstate xzs = xzstate_;
  fieldstate ys  = ystate_;
  makeState(Spectral, Spectral);

  // Looping is nonobvious. These bounds on kx avoid the kxmax cosine mode,
  // which is either not present (Nx even), or doesn't have a symmetric 
  // counterpart (Nx odd). 2007-02-23 notes.
  Complex tmp;
  for (int i=0; i<Nd_; ++i) {
    int isign = (i==2) ? 1 : -1;

    for (int ky=0; ky<Ny_; ++ky) {
      int ysign = (ky % 2 == 0) ? 1 : -1;  

      for (int kx_m = kxmin(); kx_m<=0; ++kx_m) {
    int mx_m = mx(kx_m);  // mx minus
    int mx_p = mx(-kx_m); // mx plus

    for (int mz=0; mz<Mz(); ++mz) {
      int xzsign = ((kx_m + kz(mz)) % 2 == 0) ? 1 : -1;
      Real sig = Real(isign*ysign*xzsign);

      tmp = cmplx(mx_m,ky,mz,i);
      cmplx(mx_m,ky,mz,i) = sig*cmplx(mx_p,ky,mz,i);
      cmplx(mx_p,ky,mz,i) = sig*tmp;
    }
      }
    }
  }
  makeState(xzs, ys);
  return *this;
}

FlowField& FlowField::translate(Real dx, Real dz) {
  fieldstate xzs = xzstate_;
  makeState(Spectral, ystate_);
  for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny) 
      for (int mx=0; mx<Mx(); ++mx) {
    Complex cx = exp(Complex(0.0, 2*pi*kx(mx)*dx/Lx_));
    for (int mz=0; mz<Mz(); ++mz) {
      Complex cz = exp(Complex(0.0, 2*pi*kz(mz)*dz/Lz_));
      cmplx(mx,ny,mz,i) *= cx*cz;
    }
      }
  makeState(xzs, ystate_);
  return *this;
}

// 2007-03-12 notes
FlowField& FlowField::operator *= (const FieldSymmetry& s) {
  assert(Nd_ == 3);

  // Identity escape clause
  if (s.sx()==1 && s.sy()==1 && s.sz()==1 && s.fx()==0.0 && s.fz()==0.0)
    return *this;
  
  
  fieldstate xzs = xzstate_;
  fieldstate ys  = ystate_;
  makeState(Spectral, Spectral);

  Real ax = 2*pi*s.fx();
  Real az = 2*pi*s.fz();

  Complex tmp;

  // (u,v,w)(x,y,z) -> (sa sx u, sa sy v, sa sz w) (sx x + fx*Lx, sy y, sz z + fz*Lz)
  int sa = s.sa();
  int sx = s.sx();
  int sy = s.sy();
  int sz = s.sz();

  // If one of sx,sz is -1,1 or 1,1, we must swap kx,kz and -kx,kz modes
  if (sx + sz == 0) {
    for (int i=0; i<Nd_; ++i) {
      int si = s.s(i); // (u,v,w) -> (s0 u, s1 v, s2 w)

      for (int ky=0; ky<Ny_; ++ky) {
    int syl = ((sy == -1) && (ky % 2 == 1)) ? -1 : 1;  
    Real symmsign = Real(sa*si*syl);

    for (int kx_m = kxmin(); kx_m<=0; ++kx_m) {
      int kx_p = -kx_m;
      int mx_m = mx(kx_m);  // mx minus
      int mx_p = mx(kx_p);  // mx plus
      for (int kz=0; kz<=kzmax(); ++kz) {
        int mz = this->mz(kz);
        Complex cxyz_p = symmsign*exp(Complex(0.0, ax*sx*kx_p + az*sz*kz));
        Complex cxyz_m = symmsign*exp(Complex(0.0, ax*sx*kx_m + az*sz*kz));
      
        if (sx == -1) {
          // u_{ijkl}     -> cxyz_p u_{i,-j,k,l}
          // u_{i,-j,k,l} -> cxyz_m u_{i,j,k,l}
          tmp = cmplx(mx_p,ky,mz,i);
          cmplx(mx_p,ky,mz,i) = cxyz_p*cmplx(mx_m,ky,mz,i);
          cmplx(mx_m,ky,mz,i) = cxyz_m*tmp;
        }
        else {
          // u_{ijkl}     -> cxyz_p u^*_{i,-j,k,l})
          // u_{i,-j,k,l} -> cxyz_m u^*_{i,j,k,l}
          tmp = cmplx(mx_p,ky,mz,i);
          cmplx(mx_p,ky,mz,i) = cxyz_p*conj(cmplx(mx_m,ky,mz,i));
          cmplx(mx_m,ky,mz,i) = cxyz_m*conj(tmp);
        }
        
      }
    }
      }
    }
  }
  // If sx,sz are 1,1 or -1,-1, can apply symmetry by *=
  else {
    for (int i=0; i<Nd_; ++i) {
      int si = s.sign(i); // (u,v,w) -> (s0 u, s1 v, s2 w)

      for (int ky=0; ky<Ny_; ++ky) {
    int syl = (sy == -1 && (ky % 2 == 1)) ? -1 : 1;  
    Real symmsign = Real(sa*si*syl);

    for (int kx=kxmin(); kx<=kxmax(); ++kx) {
      int mx = this->mx(kx);
      for (int kz=kzmin(); kz<=kzmax(); ++kz) {
        int mz = this->mz(kz);
        Complex cxyz = symmsign*exp(Complex(0.0,  ax*sx*kx + az*sz*kz));
      
        if (sx == 1)
          cmplx(mx,ky,mz,i) *= cxyz;
        else 
          cmplx(mx,ky,mz,i) = cxyz*conj(cmplx(mx,ky,mz,i));
      }
    }
      }
    }
  }
  makeState(xzs, ys);
  return *this;
}

bool FlowField::geomCongruent(const FlowField& v) const {
  return (Nx_ == v.Nx_ &&
      Ny_ == v.Ny_ &&
      Nz_ == v.Nz_ &&
      Lx_ == v.Lx_ &&
      Lz_ == v.Lz_ &&
      a_ == v.a_ &&
      b_ == v.b_);
}

bool FlowField::congruent(const FlowField& v) const {
  return (Nx_ == v.Nx_ &&
      Ny_ == v.Ny_ &&
      Nz_ == v.Nz_ &&
      Nd_ == v.Nd_ &&
      Lx_ == v.Lx_ &&
      Lz_ == v.Lz_ &&
      a_ == v.a_ &&
      b_ == v.b_ &&
      xzstate_ == v.xzstate_ &&
      ystate_ == v.ystate_);
}

bool FlowField::congruent(const BasisFunc& f) const {
  return (Ny_==f.Ny() && Lx_==f.Lx() && Lz_==f.Lz() &&
      a_==f.a() && b_==f.b() && ystate_==f.state());
}


FlowField& FlowField::operator *= (Real x) {
  int Ntotal = Nx_ * Ny_ * Nzpad_ * Nd_;
  for (int i=0; i<Ntotal; ++i)
    rdata_[i] *= x;
  return *this;
}

FlowField& FlowField::operator *= (Complex z) {
  assert(xzstate_ == Spectral);
  int Ntotal = Nx_ * Ny_ * Nzpad2_ * Nd_;
  for (int i=0; i<Ntotal; ++i)
    cdata_[i] *= z;
  return *this;
}

FlowField& FlowField::operator += (const ChebyCoeff& U) {
  assert(xzstate_== Spectral);
  assert(ystate_ == U.state());
  assert(Ny_ == U.N());
  for (int ny=0; ny<Ny_; ++ny) 
    cmplx(0, ny, 0, 0) += Complex(U[ny], 0.0);
  return *this;
}

FlowField& FlowField::operator -= (const ChebyCoeff& U) {
  assert(xzstate_== Spectral);
  assert(ystate_ == U.state());
  assert(Ny_ == U.N());
  for (int ny=0; ny<Ny_; ++ny) 
    cmplx(0, ny, 0, 0) -= Complex(U[ny], 0.0);
  return *this;
}

FlowField& FlowField::operator += (const ComplexChebyCoeff& U) {
  assert(xzstate_== Spectral);
  assert(ystate_ == U.state());
  assert(Ny_ == U.N());
  for (int ny=0; ny<Ny_; ++ny) 
    cmplx(0, ny, 0, 0) += U[ny];
  return *this;
}

FlowField& FlowField::operator -= (const ComplexChebyCoeff& U) {
  assert(xzstate_== Spectral);
  assert(ystate_ == U.state());
  assert(Ny_ == U.N());
  for (int ny=0; ny<Ny_; ++ny) 
    cmplx(0, ny, 0, 0) -= U[ny];
  return *this;
}

FlowField& FlowField::operator += (const BasisFunc& f){
  assert(xzstate_== Spectral);
  assert(ystate_ == f.state());
  assert(Nd_ == f.Nd());
  int m_x = mx(f.kx());
  int m_z = mz(f.kz());
  for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny) 
      cmplx(m_x, ny, m_z, i) += f[i][ny];
  return *this;
}

FlowField& FlowField::operator -= (const BasisFunc& f){
  assert(xzstate_== Spectral);
  assert(ystate_ == f.state());
  assert(Nd_ == f.Nd());
  int m_x = mx(f.kx());
  int m_z = mz(f.kz());
  for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny) 
      cmplx(m_x, ny, m_z, i) -= f[i][ny];
  return *this;
}

FlowField& FlowField::operator += (const RealProfile& f){
  assert(xzstate_== Spectral);
  assert(ystate_ == f.state());
  assert(Nd_ == f.Nd());
  int fkx = f.kx();
  int fkz = f.kz();

  if (fkx < kxmin() || fkx > kxmax() || fkz < kzmin() || fkz > kzmax()) 
    return *this;
  
  int m_x = mx(fkx);
  int m_z = mz(fkz);
  Complex minus_i(0.0, 1.0);
  const BasisFunc& psi = f.psi;

  // f = (psi + psi^*)   case Plus
  // f = (psi - psi^*)/i case Minus
  // 
  // where psi.kz >= 0 
  //       psi.kx is negative, zero, or positive.

  // Because u is real, u(-kx,-kz) = u^*(kx,kz)
  // or equivalently    u( kx,-kz) = u^*(-kx,kz)

  // Because of this symmetry and the way FFTW works, FlowField stores 
  // only u(kx,kz) for kz > 0, u(kx,-kz) is obtained from the symmetry.
  // But for kz=0 and most kx, both u(kx,0) and u(-kx,0) are stored.
  // These are handled by the if statement below
  switch (f.sign()) {

  case Minus:
    for (int i=0; i<Nd_; ++i)
      for (int ny=0; ny<Ny_; ++ny) 
    cmplx(m_x, ny, m_z, i) += minus_i*psi[i][ny];
    break;

  case Plus:
    for (int i=0; i<Nd_; ++i)
      for (int ny=0; ny<Ny_; ++ny) 
    cmplx(m_x, ny, m_z, i) += psi[i][ny];
    break;
  }

  // When kz=0, both u(kx,0) and u(-kx,0) are stored in the flow field,
  // except for the very largest kx>0 when Nx is even. So to we need to 
  // add u(-kx,0) = u^*(kx,0) for kx>1.
  if (fkz==0 && fkx>=0 && -fkx >= kxmin()) {
    int nmx = mx(-fkx);
    switch (f.sign()) {

    case Minus:
      for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny) 
      cmplx(nmx, ny, m_z, i) += conj(minus_i*(psi[i][ny]));
      break;

    case Plus:
      for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny) 
      cmplx(nmx, ny, m_z, i) += conj(psi[i][ny]);
      break;
    }
  }
  return *this;
}

FlowField& FlowField::operator -= (const RealProfile& f){
  assert(xzstate_== Spectral);
  assert(ystate_ == f.state());
  assert(Nd_ == f.Nd());

  int fkx = f.kx();
  int fkz = f.kz();

  if (fkx < kxmin() || fkx > kxmax() || fkz < kzmin() || fkz > kzmax())
    return *this;

  int m_x = mx(fkx);
  int m_z = mz(fkz);
  Complex minus_i(0.0, 1.0);
  const BasisFunc& psi = f.psi;

  switch (f.sign()) {

  case Minus:
    for (int i=0; i<Nd_; ++i)
      for (int ny=0; ny<Ny_; ++ny) 
    cmplx(m_x, ny, m_z, i) -= minus_i*psi[i][ny];
    break;

  case Plus:
    for (int i=0; i<Nd_; ++i)
      for (int ny=0; ny<Ny_; ++ny) 
    cmplx(m_x, ny, m_z, i) -= psi[i][ny];
    break;
  }

  if (fkz==0 && fkx>=0 && -fkx >= kxmin()) {
    int nmx = mx(-fkx);
    switch (f.sign()) {

    case Minus:
      for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny) 
      cmplx(nmx, ny, m_z, i) -= conj(minus_i*(psi[i][ny]));
      break;

    case Plus:
      for (int i=0; i<Nd_; ++i)
    for (int ny=0; ny<Ny_; ++ny) 
      cmplx(nmx, ny, m_z, i) -= conj(psi[i][ny]);
      break;
    }
  }
  return *this;
}


FlowField& FlowField::operator += (const FlowField& u) {
  assert(congruent(u));
  int Ntotal = Nx_ * Ny_ * Nzpad_ * Nd_;
  for (int i=0; i<Ntotal; ++i)
    rdata_[i] += u.rdata_[i];
  return *this;
}
FlowField& FlowField::operator -= (const FlowField& u) {
  assert(congruent(u));
  int Ntotal = Nx_ * Ny_ * Nzpad_ * Nd_;
  for (int i=0; i<Ntotal; ++i)
    rdata_[i] -= u.rdata_[i];
  return *this;
}

FlowField  FlowField::operator[](int i) const {
  FlowField ui(Nx_, Ny_, Nz_, 1, Lx_, Lz_, a_, b_, xzstate_, ystate_); 
  
  if (xzstate_ == Spectral) 
    for (int my=0; my<My(); ++my)
      for (int mx=0; mx<Mx(); ++mx)
    for (int mz=0; mz<Mz(); ++mz)
      ui.cmplx(mx,my,mz,0) = this->cmplx(mx,my,mz,i);
  else
    for (int ny=0; ny<Ny(); ++ny)
      for (int nx=0; nx<Nx(); ++nx)
    for (int nz=0; nz<Nz(); ++nz)
      ui(nx,ny,nz,0) = (*this)(nx,ny,nz,i);
  return ui;
}

void FlowField::realfft_xz() {
  assert (xzstate_ == Physical);
  fftw_execute(xz_plan_);
  // args are (plan, howmany, in, istride, idist, ostride, odist)
  //rfftwnd_real_to_complex(xz_plan_, Nd_*Ny_, rdata_, 1, Nx_*Nzpad_, 0,0,0);
  int Ntotal = Nd_* Nx_ * Ny_ * Nzpad_;
  Real scale = 1.0/(Nx_*Nz_);
  for (int i=0; i<Ntotal; ++i)
    rdata_[i] *= scale;
  xzstate_ = Spectral;
}

void FlowField::irealfft_xz() {
  assert(xzstate_ == Spectral);
  //cout << "rfftwnd_complex_to_real : " << endl;
  //cout << "howmany = " << Nd_*Ny_ << endl;
  //cout << "idist = " << Nx_*Nzpad_/2 << endl;

  fftw_execute(xz_iplan_);
  //rfftwnd_complex_to_real(xz_iplan_, Nd_*Ny_, (fftw_complex*)rdata_, 1, Nx_*Nzpad_/2, 0,0,0);
  xzstate_ = Physical;
}

void FlowField::chebyfft_y() {
  assert(ystate_ == Physical);
  if (Ny_ < 2) {
    ystate_ = Spectral;
    return;
  }
    
  Real nrm = 1.0/(Ny_-1); // needed because FFTW does unnormalized transforms
  for (int i=0; i<Nd_; ++i) 
    for (int nx=0; nx<Nx_; ++nx) 
      for (int nz=0; nz<Nzpad_; ++nz) {

    // Copy data spread through memory into a stride-1 scratch array.
    for (int ny=0; ny<Ny_; ++ny) 
      scratch_[ny] = rdata_[flatten(nx, ny, nz, i)];

    // Transform data
    fftw_execute_r2r(y_plan_, scratch_, scratch_);

    //for (int ny=2*(Ny_-1)/3+1; ny<Ny_; ++ny) scratch_[ny] = 0.0; //dealiasing for y-mode

    // Copy back to multi-d arrays, normalizing on the way
    // 0th elem is different because of reln btwn cos and Cheb transforms
    rdata_[flatten(nx, 0, nz, i)] = 0.5*nrm*scratch_[0]; 
    for (int ny=1; ny<Ny_-1; ++ny) 
      rdata_[flatten(nx, ny, nz, i)] = nrm*scratch_[ny];
    rdata_[flatten(nx, Ny_-1, nz, i)] = 0.5*nrm*scratch_[Ny_-1];
      }
  ystate_ = Spectral;
  return;
}

void FlowField::ichebyfft_y() {
  assert(ystate_ == Spectral);
  if (Ny_ < 2) {
    ystate_ = Physical;
    return;
  }
  for (int i=0; i<Nd_; ++i) 
    for (int nx=0; nx<Nx_; ++nx) 
      for (int nz=0; nz<Nzpad_; ++nz) {

    // Copy data spread through memory into a stride-1 scratch array.
    // 0th and last elems are different because of reln btwn cos and 
    // Cheb transforms.
    scratch_[0] = rdata_[flatten(nx, 0, nz, i)];
    for (int ny=1; ny<Ny_-1; ++ny) 
      scratch_[ny] = 0.5*rdata_[flatten(nx, ny, nz, i)];
    scratch_[Ny_-1] = rdata_[flatten(nx, Ny_-1, nz, i)];

    // Transform data
    fftw_execute_r2r(y_plan_, scratch_, scratch_);

    // Copy transformed data back into main data array
    for (int ny=0; ny<Ny_; ++ny) 
      rdata_[flatten(nx, ny, nz, i)] = scratch_[ny];

      }
  ystate_ = Physical;
  return;
}

void FlowField::makeSpectral_xz() {if (xzstate_==Physical) realfft_xz();}
void FlowField::makePhysical_xz() {if (xzstate_==Spectral) irealfft_xz();}
void FlowField::makeSpectral_y()  {if (ystate_==Physical) chebyfft_y();}
void FlowField::makePhysical_y()  {if (ystate_==Spectral) ichebyfft_y();}
void FlowField::makeSpectral()  {makeSpectral_xz(); makeSpectral_y();}
void FlowField::makePhysical()  {makePhysical_y(); makePhysical_xz();}
void FlowField::makeState(fieldstate xzstate, fieldstate ystate)  {
  (xzstate == Physical) ? makePhysical_xz() : makeSpectral_xz();
  (ystate  == Physical) ? makePhysical_y()  : makeSpectral_y();
}

Complex FlowField::Dx(int mx, int n) const {
  Complex rot(0.0, 0.0);
  switch(n % 4) {
  case 0: rot = Complex( 1.0, 0.0); break;
  case 1: rot = Complex( 0.0, 1.0); break;
  case 2: rot = Complex(-1.0, 0.0); break;
  case 3: rot = Complex( 0.0,-1.0); break;
  default: cferror("FlowField::Dx(mx,n) : impossible: n % 4 > 4 !!");
  }
  int kx_ = kx(mx);
  return rot*(std::pow(2*pi*kx_/Lx_, n)*zero_last_mode(kx_,kxmax(),n));
}
Complex FlowField::Dz(int mz, int n) const {
  Complex rot(0.0, 0.0);
  switch(n % 4) {
  case 0: rot = Complex( 1.0, 0.0); break;
  case 1: rot = Complex( 0.0, 1.0); break;
  case 2: rot = Complex(-1.0, 0.0); break;
  case 3: rot = Complex( 0.0,-1.0); break;
  default: cferror("FlowField::Dx(mx,n) : impossible: n % 4 > 4 !!");
  }
  int kz_ = kz(mz);
  return rot*(std::pow(2*pi*kz_/Lz_, n)*zero_last_mode(kz_,kzmax(),n));
}

Complex FlowField::Dx(int mx) const {
  int kx_ = kx(mx);
  return Complex(0.0, 2*pi*kx_/Lx_*zero_last_mode(kx_,kxmax(),1));
}
Complex FlowField::Dz(int mz) const {
  int kz_ = kz(mz);
  return Complex(0.0, 2*pi*kz_/Lz_*zero_last_mode(kz_,kzmax(),1));
}


void FlowField::addPerturbation(int kx, int kz, Real mag, Real decay) {
  assertState(Spectral, Spectral);
  if (mag == 0.0)
    return;

  // Add a dive-free perturbation to the base flow.
  ComplexChebyCoeff u(Ny_, a_, b_, Spectral);
  ComplexChebyCoeff v(Ny_, a_, b_, Spectral);
  ComplexChebyCoeff w(Ny_, a_, b_, Spectral);
  randomProfile(u,v,w, kx, kz, Lx_, Lz_, mag, decay);
  int m_x = mx(kx);
  int m_z = mz(kz);
  for (int ny=0; ny<Ny_; ++ny) {
    cmplx(m_x, ny, m_z, 0) += u[ny]; //to x-direction
    cmplx(m_x, ny, m_z, 1) += v[ny]; //to y-direction
    cmplx(m_x, ny, m_z, 2) += w[ny]; //to z-direction
  }
  if (kz==0 && kx!=0) {
    int m_x = mx(-kx);
    int m_z = mz(0); // -kz=0
    for (int i=0; i<Nd_; ++i)
      for (int ny=0; ny<Ny_; ++ny) {
    cmplx(m_x, ny, m_z, 0) += conj(u[ny]); //to x-direction
    cmplx(m_x, ny, m_z, 1) += conj(v[ny]); //to y-direction
    cmplx(m_x, ny, m_z, 2) += conj(w[ny]); //to z-direction
      }
  }
  return;
}

void FlowField::perturb(Real mag, Real decay) {
  addPerturbations(kxmax(), kzmax(), mag, decay);
}

void FlowField::addPerturbations(int Kx, int Kz, Real mag, Real decay) {
  assertState(Spectral, Spectral);
  if (mag == 0.0)
    return;

  int Kxmin = Greater(-Kx, kxmin());
  int Kxmax = lesser(Kx, kxmax()-1);
  int Kzmax = lesser(Kz, kzmax()-1);

  // Add a div-free perturbation to the base flow.
  for (int kx=Kxmin; kx<=Kxmax; ++kx) 
    for (int kz=0; kz<=Kzmax; ++kz) {
      //Real norm = pow(10.0, -(abs(2*pi*kx/Lx_) + abs(2*pi*kz/Lz_)));
      //Real norm = pow(decay, 2*(abs(kx) + abs(kz)));
      Real norm = std::pow(decay, abs(2*pi*kx/Lx_) + abs(2*pi*kz/Lz_));
      if (!(kx==0 && kz==0))
    addPerturbation(kx, kz, mag*norm, decay);
    }
  makePhysical();
  makeSpectral();
  return;
}

void FlowField::addThePerturbations(int Kx, int Kz, Real mag, Real decay) {
  assertState(Spectral, Spectral);
  if (mag == 0.0)
    return;

  //int Kxmin = Greater(-Kx, kxmin());
  int Kxmax = lesser(Kx, kxmax()-1);
  int Kzmax = lesser(Kz, kzmax()-1);

  // Add a div-free perturbation to the base flow.
  for (int kx=Kxmax-1; kx<=Kxmax; ++kx) 
    for (int kz=Kzmax-1; kz<=Kzmax; ++kz) {
      //Real norm = pow(10.0, -(abs(2*pi*kx/Lx_) + abs(2*pi*kz/Lz_)));
      //Real norm = pow(decay, 2*(abs(kx) + abs(kz)));
      Real norm = std::pow(decay, abs(2*pi*kx/Lx_) + abs(2*pi*kz/Lz_));
      if (!(kx==0 && kz==0))
    addPerturbation(kx, kz, mag*norm, decay);
    }
  makePhysical();
  makeSpectral();
  return;
}

void FlowField::addThePerturbations(Real K, Real mag, Real decay) {
  assertState(Spectral, Spectral);
  if (mag <= 0.0) return;
  decay = abs(decay);

  int Kxmin = kxminDealiased();
  int Kxmax = kxmaxDealiased();
  int Kzmin = kzminDealiased();
  int Kzmax = kzmaxDealiased();
  Real KK;

  // Add a div-free perturbation to the base flow.
  for (int kx=Kxmin; kx<=Kxmax; ++kx) 
    for (int kz=Kzmin; kz<=Kzmax; ++kz) {
      //Real norm = pow(10.0, -(abs(2*pi*kx/Lx_) + abs(2*pi*kz/Lz_)));
      //Real norm = pow(decay, 2*(abs(kx) + abs(kz)));
      KK = sqrt(((double)kx*kx+kz*kz));
      if (!(kx==0 && kz==0) && (KK < K+decay && KK > K-decay)) {
      Real norm = std::pow(decay, abs(2*pi*kx/Lx_) + abs(2*pi*kz/Lz_));
      addPerturbation(kx, kz, mag*norm, decay);
      }
    }
  makePhysical();
  makeSpectral();
  return;
}

void FlowField::addPerturbations(Real mag, Real decay) {
  assertState(Spectral, Spectral);
  if (mag == 0.0)
    return;

  // Add a div-free perturbation to the base flow.
  for (int mx=0; mx<Mx(); ++mx) {
    int kx_ = kx(mx);
    for (int mz=0; mz<Mz(); ++mz) {
      int kz_ = kz(mz);
      Real norm = std::pow(decay, 2*(abs(kx_) + abs(kz_)));
      addPerturbation(kx_, kz_, mag*norm, decay);
    }
  }
  makePhysical();
  makeSpectral();
  return;
}
bool FlowField::dealiased() const { return dealiased_;}
  
void FlowField::setDealiased(bool b) {dealiased_ = b;}

void FlowField::setToZero() {
  int Ntotal = Nx_*Ny_*Nzpad_*Nd_;
  for (int i=0; i<Ntotal; ++i)
    rdata_[i] = 0.0;
}

void FlowField::zeroAliasedModes() {
  fieldstate xzs = xzstate_;
  fieldstate ys =  ystate_;
  makeSpectral_xz();

  // Not as efficient as possible but less prone to error and more robust 
  // to changes in aliasing bounds, and the efficiency diff is negligible
  Complex zero(0.0, 0.0);
  for (int mx=0; mx<Mx(); ++mx) {
    int kx_ = this->kx(mx);
    for (int mz=0; mz<Mz(); ++mz) {
      int kz_ = this->kz(mz);
      if (isAliased(kx_,kz_))
    for (int i=0; i<Nd_; ++i)
      for (int ny=0; ny<Ny_; ++ny)
        this->cmplx(mx,ny,mz,i) = zero;
    }
  }
  dealiased_ = true;
  makeState(xzs,ys);
}

void FlowField::print() const {
  cout << Nx_ << " x " << Ny_ << " x " << Nz_ << endl;
  cout << "[0, " << Lx_ << "] x [-1, 1] x [0, " << Lz_ << "]" << endl;
  cout << xzstate_ << " x " << ystate_ << " x " << xzstate_ << endl;
  cout << xzstate_ << " x " << ystate_ << " x " << xzstate_ << endl;
  if (xzstate_ == Spectral) {
  cout << "FlowField::print() real view " << endl;
    for (int i=0; i<Nd_; ++i) {
      for (int ny=0; ny<Ny_; ++ny) {
    for (int nx=0; nx<Nx_; ++nx) {
       cout << "i=" << i << " ny=" << ny << " nx= " << nx << ' ';
       int nz; // MSVC++ FOR-SCOPE BUG
      for (nz=0; nz<Nz_; ++nz) 
        cout << rdata_[flatten(nx,ny,nz,i)] << ' ';
      cout << " pad : ";
      for (nz=Nz_; nz<Nzpad_; ++nz) 
        cout << rdata_[flatten(nx,ny,nz,i)] << ' ';
      cout << endl;
    }
      }
    }
  }
  else {
    cout << "complex view " << endl;
    for (int i=0; i<Nd_; ++i) {
       for (int ny=0; ny<Ny_; ++ny) {
     for (int nx=0; nx<Nx_; ++nx) {
       cout << "i=" << i << " ny=" << ny << " nx= " << nx << ' ';
       for (int nz=0; nz<Nz_/2; ++nz) 
         cout << cdata_[complex_flatten(nx,ny,nz,i)] << ' ';
       cout << endl;
     }
       }    
    }
  }
}

// k == direction of normal    (e.g. k=0 means a x-normal slice in yz plane.
// i == component of FlowField (e.g. i=0 means u-component)
// n == nth gridpoint along k direction
void FlowField::saveSlice(int k, int i, int nk, const string& filebase) const {
  assert(k>=0 && k<3);
  assert(i>=0 && i<Nd_);


  string filename(filebase);
  filename += string(".asc");
  ofstream os(filename.c_str());

  FlowField& u = (FlowField&) *this;  // cast away constness
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

  u.makePhysical();

  switch (k) {
  case 0: 
    os << "% yz slice\n";
    os << "% (i,j)th elem is field at (x_n, y_i, z_j)\n";
    for (int ny=0; ny<Ny_; ++ny) {
      for (int nz=0; nz<Nz_; ++nz) 
    os << u(nk, ny, nz, i) << ' ';
      os << u(nk, ny, 0, i) << '\n';  // repeat nz=0 to complete [0, Lz]
    }
    break;
  case 1: // xz slice
    os << "% xz slice\n";
    os << "% (i,j)th elem is field at (x_j, y_n, z_i)\n";
    for (int nz=0; nz<Nz_; ++nz) {
      for (int nx=0; nx<Nx_; ++nx) 
    os << u(nx, nk, nz, i) << ' ';
      os << u(0, nk, nz, i) << '\n'; // repeat nx=0 to complete [0, Lx]
    }
    for (int nx=0; nx<Nx_; ++nx)     // repeat nz=0 to complete [0, Lz]
      os << u(nx, nk, 0, i) << ' ';
    os << u(0, nk, 0, i) << '\n';    // repeat nx=0 to complete [0, Lx]

    break;
  case 2: 
    os << "% yz slice\n";
    os << "% (i,j)th elem is field at (x_j, y_i, z_n)\n";
    for (int ny=0; ny<Ny_; ++ny) {
      for (int nx=0; nx<Nx_; ++nx) 
    os << u(nx, ny, nk, i) << ' ';
      os << u(0, ny, nk, i) << '\n';
    }
    break;
  }
  u.makeState(xzstate, ystate);
}

void FlowField::saveProfile(int mx, int mz, const string& filebase) const {
  ChebyTransform trans(Ny_);
  saveProfile(mx,mz,filebase, trans);
}

void FlowField::saveProfile(int mx, int mz, const string& filebase, const ChebyTransform& trans) const {
  assert(xzstate_ == Spectral);
  string filename(filebase);
  if (Nd_ == 3)
    filename += string(".bf");  // this convention is unfortunate, need to fix
  else 
    filename += string(".asc");

  ofstream os(filename.c_str());
  os << setprecision(REAL_DIGITS);

  if (ystate_ == Physical) {
    for (int ny=0; ny<Ny_; ++ny) {
      for (int i=0; i<Nd_; ++i) {
    Complex c = (*this).cmplx(mx, ny, mz, i);
    os << Re(c) << ' ' << Im(c) << ' ';
      }
      os << '\n';
    }
  }
  else {
    ComplexChebyCoeff* f = new ComplexChebyCoeff[Nd_];
    for (int i=0; i<Nd_; ++i) {
      f[i] = ComplexChebyCoeff(Ny_, a_, b_, Spectral);
      for (int ny=0; ny<Ny_; ++ny) 
    f[i].set(ny, (*this).cmplx(mx, ny, mz, i));
      f[i].makePhysical(trans);
    }
    for (int ny=0; ny<Ny_; ++ny) {
      for (int i=0; i<Nd_; ++i) 
    os << f[i].re[ny] << ' ' << f[i].im[ny] << ' ';
      os << '\n';
    }
  }
}

void FlowField::AscSave(const std::string& filebase) {

  FlowField& u = (FlowField&) *this;  // cast away constness
 // backup current state
  fieldstate uxzstate = xzstate_;
  fieldstate uystate = ystate_;

  int ix, iy, iz, idir;

  string filename = filebase;
  FILE *fpu;
  fpu = fopen(filename.c_str(), "w");
  if (fpu == NULL) {
    cerr << "FlowField::AscSave can't open file " << endl;
    abort();
  }

  u.makePhysical();
  iy = 0; idir = 0;
  int iy1, iy2, idir1, idir2;
  iy1 = 0; iy2 = Ny_-1;
  idir1 = 0; idir2 = 2;
  for(ix = 0; ix<Nx_; ix++) {
    for(iz = 0; iz<Nz_; iz++) {
      fprintf(fpu, "%14.7e %14.7e %14.7e %14.7e %14.7e %14.7e\n",u.x(ix), u.z(iz), 
        u(ix,iy1,iz,idir1), u(ix,iy2,iz,idir1), u(ix,iy1,iz,idir2), u(ix,iy2,iz,idir2));
    }
  }
  
  fclose(fpu);

// recovery current state  
  u.makeState(uxzstate, uystate);

}

void FlowField::saveSpectrum(const string& filebase, Real nu, int iy) const {
  string filename(filebase);
  //filename += string(".asc");
  ofstream os(filename.c_str());

  assert(xzstate_ == Spectral);
  
  Real rkk, rk, rkmax, tmp, factor;
  rkmax = sqrt(((double)kxmaxDealiased()*kxmaxDealiased()+kzmaxDealiased()*kzmaxDealiased()));
  
  Real spectrum[int(rkmax)+1][2];
  
  int i;
  int Kxmin = kxminDealiased();
  int Kxmax = kxmaxDealiased();
  int Kzmin = kzminDealiased();
  int Kzmax = kzmaxDealiased();
    
  for (i=0; i<int(rkmax)+1; ++i) {
  spectrum[i][0] = 0.0; spectrum[i][1] = 0.0;
  }
  
  factor = 1.0;
  for (int kx=Kxmin; kx<=Kxmax; ++kx) {
    for (int kz=Kzmin; kz<=Kzmax; ++kz) {
    rkk = kx*kx+kz*kz; rk = sqrt(rkk); 
    //tmp = factor*sqrt(energy(mx(kx),mz(kz)));
    tmp = factor*energy(mx(kx),mz(kz),iy);
    spectrum[int(rk)][0] += tmp;
    spectrum[int(rk)][1] += 2.0*rkk*tmp;
    }
  }

  rk = (kxmaxDealiased() > kzmaxDealiased()) ? kzmaxDealiased():kxmaxDealiased();
  Real K, D, ReL, Re_lambda, eta;
  K = 0.0; D = 0.0;
  for (i=0; i<int(rk)+2; ++i) {
  K +=spectrum[i][0]; D +=spectrum[i][1];
  }
  ReL=(K-spectrum[0][0])*(K-spectrum[0][0])/(D*nu);
  Re_lambda= sqrt(20.0*ReL/3.0);
  eta=sqrt(sqrt(nu*nu*nu/D));
  
  Real efactor = 1.0/sqrt(sqrt(D*nu*nu*nu*nu*nu));
  os<<"# 3D Spectrum, K="<<K<<", Re_lambda="<<Re_lambda<<", Re_L="<<ReL<<", eta="<<eta<<", D="<<D<<endl;
  for (i=0; i<int(rk)+2; ++i) {
  os << i <<"  "<< spectrum[i][0] <<"  "<< spectrum[i][1];
  os <<"  "<< i*eta <<"  "<< spectrum[i][0]*efactor << endl;
  }
  os.close();
}

void FlowField::saveSpectrum(const string& filebase, Real nu) const {
  string filename(filebase);
  //filename += string(".asc");
  ofstream os(filename.c_str());

  assert(xzstate_ == Spectral);
  
  Real rkk, rk, rkmax, tmp, factor;
  rkmax = sqrt(((double)kxmaxDealiased()*kxmaxDealiased()+kzmaxDealiased()*kzmaxDealiased()));
  
  Real spectrum[int(rkmax)+1][2];
  
  int i;
  int Kxmin = kxminDealiased();
  int Kxmax = kxmaxDealiased();
  int Kzmin = kzminDealiased();
  int Kzmax = kzmaxDealiased();
    
  for (i=0; i<int(rkmax)+1; ++i) {
  spectrum[i][0] = 0.0; spectrum[i][1] = 0.0;
  }
  
  factor = 1.0;
  for (int kx=Kxmin; kx<=Kxmax; ++kx) {
    for (int kz=Kzmin; kz<=Kzmax; ++kz) {
    rkk = kx*kx+kz*kz; rk = sqrt(rkk); 
    //tmp = factor*sqrt(energy(mx(kx),mz(kz)));
    tmp = factor*energy(mx(kx),mz(kz));
    spectrum[int(rk)][0] += tmp;
    spectrum[int(rk)][1] += 2.0*rkk*tmp;
    }
  }

  rk = (kxmaxDealiased() > kzmaxDealiased()) ? kzmaxDealiased():kxmaxDealiased();
  Real K, D, ReL, Re_lambda, eta;
  K = 0.0; D = 0.0;
  for (i=0; i<int(rk)+2; ++i) {
  K +=spectrum[i][0]; D +=spectrum[i][1];
  }
  ReL=(K-spectrum[0][0])*(K-spectrum[0][0])/(D*nu);
  Re_lambda= sqrt(20.0*ReL/3.0);
  eta=sqrt(sqrt(nu*nu*nu/D));
  
  os<<"# 3D Spectrum, K="<<K<<", Re_lambda="<<Re_lambda<<", Re_L="<<ReL<<", eta="<<eta<<", D="<<D<<endl;
  for (i=0; i<int(rk)+2; ++i) {
  os << i <<"  "<< spectrum[i][0] <<"  "<< spectrum[i][1] << endl;
  }
  os.close();
}

void FlowField::saveSpectrum(const string& filebase, int i, int ny, 
                 bool kxorder, bool showpadding) const {
  string filename(filebase);
  filename += string(".asc");
  ofstream os(filename.c_str());

  bool sum = (ny == -1) ? true : false;
  assert(xzstate_ == Spectral);
  
  if (kxorder) {
    int Kxmin = (showpadding || !dealiased_) ? kxmin() : kxminDealiased();
    int Kxmax = (showpadding || !dealiased_) ? kxmax() : kxmaxDealiased();
    int Kzmin = (showpadding || !dealiased_) ? kzmin() : kzminDealiased();
    int Kzmax = (showpadding || !dealiased_) ? kzmax() : kzmaxDealiased();
    for (int kx=Kxmin; kx<=Kxmax; ++kx) {
      for (int kz=Kzmin; kz<=Kzmax; ++kz) {
    if (sum) 
      os << sqrt(energy(mx(kx),mz(kz))) << ' ';
    else {
      Complex f = this->cmplx(mx(kx), ny, mz(kz), i);
      os << Re(f) << ' ' << Im(f) << ' ';
    }
      }
      os << endl;
    }
  }
  else {
    for (int mx=0; mx<Mx(); ++mx) {
      for (int mz=0; mz<Mz(); ++mz) {
    if (sum) 
      os << sqrt(energy(mx,mz)) << ' ';
    else {
      Complex f = this->cmplx(mx, ny, mz, i);
      os << Re(f) << ' ' << Im(f) << ' ';
    }
      }
      os << endl;
    }
  }
  os.close();
}

void FlowField::saveSpectrum(const string& filebase, bool kxorder, 
                 bool showpadding) const {
  string filename(filebase);
  filename += string(".asc");
  ofstream os(filename.c_str());

  assert(xzstate_ == Spectral && ystate_ == Spectral);
 
  ComplexChebyCoeff u(Ny_,a_,b_,Spectral);

  if (kxorder) {
    int Kxmin = (showpadding || !dealiased_) ? kxmin() : kxminDealiased();
    int Kxmax = (showpadding || !dealiased_) ? kxmax() : kxmaxDealiased();
    int Kzmin = (showpadding || !dealiased_) ? kzmin() : kzminDealiased();
    int Kzmax = (showpadding || !dealiased_) ? kzmax() : kzmaxDealiased();
    for (int kx=Kxmin; kx<Kxmax; ++kx) {
      for (int kz=Kzmin; kz<Kzmax; ++kz) {
    Real e = 0.0;
    for (int i=0; i<Nd_; ++i) {
      for (int ny=0; ny<Ny_; ++ny) 
        u.set(ny,this->cmplx(mx(kx),ny,mz(kz),i));
      e += L2Norm2(u);
    }
    os << sqrt(e) << ' ';
      }
      os << endl;
    }
  }
  else {
    for (int mx=0; mx<Mx(); ++mx) {
      for (int mz=0; mz<Mz(); ++mz) {
    Real e = 0.0;
    for (int i=0; i<Nd_; ++i) {
      for (int ny=0; ny<Ny_; ++ny) 
        u.set(ny,this->cmplx(mx,ny,mz,i));
      e += L2Norm2(u);
    }
    os << sqrt(e) << ' ';
      }
      os << endl;
    }
  }
  os.close();
}


void FlowField::saveDivSpectrum(const string& filebase, bool kxorder,
                bool showpadding) const {
  string filename(filebase);
  filename += string(".asc");
  ofstream os(filename.c_str());

  assert(xzstate_ == Spectral && ystate_ == Spectral);
  
  //assert(congruent(div));
  assert(Nd_ >= 3);

  ComplexChebyCoeff v(Ny_, a_, b_, Spectral);
  ComplexChebyCoeff vy(Ny_, a_, b_, Spectral);

  // Rely on compiler to pull loop invariants out of loops
  if (kxorder) {
    int Kxmin = kxmin();
    int Kxmax = (showpadding || !dealiased_) ? kxmax() : kxmaxDealiased();
    int Kzmin = (showpadding || !dealiased_) ? kzmin() : kzminDealiased();
    int Kzmax = (showpadding || !dealiased_) ? kzmax() : kzmaxDealiased();
    for (int kx=Kxmin; kx<Kxmax; ++kx) {
      Complex d_dx(0.0, 2*pi*kx/Lx_*zero_last_mode(kx,kxmax(),1));

      for (int kz=Kzmin; kz<Kzmax; ++kz) {
    Complex d_dz(0.0, 2*pi*kz/Lz_*zero_last_mode(kz,kzmax(),1));

    for (int ny=0; ny<Ny_; ++ny) 
      v.set(ny, cmplx(mx(kx),ny,mz(kz),1));
    diff(v,vy);

    Real div = 0.0;
    for (int ny=0; ny<Ny_; ++ny) {
      Complex ux = d_dx*cmplx(mx(kx),ny,mz(kz),0);
      Complex wz = d_dz*cmplx(mx(kx),ny,mz(kz),2);
      div += abs2(ux + vy[ny] + wz);
    }
    os << sqrt(div) << ' ';
      }
      os << endl;
    }
  }
  else {
    for (int mx=0; mx<Mx(); ++mx) {
      Complex d_dx(0.0, 2*pi*kx(mx)/Lx_*zero_last_mode(kx(mx),kxmax(),1));
      
      for (int mz=0; mz<Mz(); ++mz) {
    Complex d_dz(0.0, 2*pi*kz(mz)/Lz_*zero_last_mode(kz(mz),kzmax(),1));
      
    for (int ny=0; ny<Ny_; ++ny) 
      v.set(ny, cmplx(mx,ny,mz,1));
    diff(v,vy);

    Real div = 0.0;
    for (int ny=0; ny<Ny_; ++ny) {
      Complex ux = d_dx*cmplx(mx,ny,mz,0);
      Complex wz = d_dz*cmplx(mx,ny,mz,2);
      div += abs2(ux + vy[ny] + wz);
    }
    os << sqrt(div) << ' ';
      }
      os << endl;
    }
  }
  os.close();
}

void FlowField::asciiSave(const string& filebase) const {
  string filename(filebase);
  filename += ".asc";

  ofstream os(filename.c_str());
  os << scientific << setprecision(REAL_DIGITS);

  const char s = ' ';
  const char nl = '\n';
  const int w = REAL_IOWIDTH;
  os << "% Channelflow FlowField data" << nl;
  os << "% xzstate == " << xzstate_ << nl;
  os << "%  ystate == " << ystate_ << nl;
  os << "% Nx Ny Nz Nd == " << Nx_ <<s<< Ny_ <<s<< Nz_ <<s<< Nd_ << " gridpoints\n";
  os << "% Mx My Mz Nd == " << Mx() <<s<< My() <<s<< Mz() <<s<< Nd_ << " spectral modes\n";
  os << "% Lx Lz == " << setw(w) << Lx_ <<s<< setw(w) << Lz_ << nl;
  os << "% Lx Lz == " << setw(w) << Lx_ <<s<< setw(w) << Lz_ << nl;
  os << "% a  b  == " << setw(w) << a_ <<s<< setw(w) << b_ << nl;
  os << "% loop order:\n";
  os << "%   for (int i=0; i<Nd; ++i)\n";
  os << "%     for(long ny=0; ny<Ny; ++ny) // note: Ny == My\n";
  if (xzstate_ == Physical) {
    os << "%       for (int nx=0; nx<Nx; ++nx)\n";
    os << "%         for (int nz=0; nz<Nz; ++nz)\n";
    os << "%           os << f(nx, ny, nz, i) << newline;\n";
  } 
  else {
    os << "%       for (int mx=0; mx<Mx; ++mx)\n";
    os << "%         for (int mz=0; mz<Mz; ++mz)\n";
    os << "%           os << Re(f.cmplx(mx, ny, mz, i) << ' ' << Im(f.cmplx(mx, ny, mz, i) << newline;\n";
  }
  if (xzstate_ == Physical) {
    for (int i=0; i<Nd_; ++i) 
      for(long ny=0; ny<Ny_; ++ny)
    for (int nx=0; nx<Nx_; ++nx)
      for (int nz=0; nz<Nz_; ++nz)
        os << setw(w) << (*this)(nx, ny, nz, i) << nl;
  }
  else {
    for (int i=0; i<Nd_; ++i) 
      for(long ny=0; ny<Ny_; ++ny)
    for (int mx=0; mx<Mx(); ++mx)
      for (int mz=0; mz<Mz(); ++mz) 
        os << setw(w) << Re(cmplx(mx, ny, mz, i)) << s
           << setw(w) << Im(cmplx(mx, ny, mz, i)) << nl;
  }
  os.close();
}
      
void FlowField::binarySave(const string& filebase) const {

  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".ff")
    filename = filebase;
  else
    filename += string(".ff"); // "flow field"
  *******************************/
  string filename = appendSuffix(filebase, ".ff");

  ofstream os(filename.c_str(), ios::out | ios::binary);
  if (!os.good()) {
    cerr << "FlowField::binarySave(filebase) : can't open file " << filename << endl;
    abort();
  }

  int major; 
  int minor;
  int update;
  channelflowVersion(major, minor, update);

  write(os, major);
  write(os, minor);
  write(os, update);
  write(os, Nx_);
  write(os, Ny_);
  write(os, Nz_);
  write(os, Nd_);
  write(os, xzstate_);
  write(os, ystate_);
  write(os, Lx_);
  write(os, Lz_);
  write(os, a_);
  write(os, b_);
  write(os, dealiased_);

  // Write data only for non-aliased modes.
  if (dealiased_ && xzstate_ == Spectral) {
    int Nxd=2*(Nx_/6);
    int Nzd=2*(Nz_/3)+1;

    // In innermost loop, array index is (nz + Nzpad2_*(nx + Nx_*(ny + Ny_*i))),
    // which is the same as the FlowField::flatten function.
    for (int i=0; i<Nd_; ++i) {
      for (int ny=0; ny<Ny_; ++ny) {

    for (int nx=0; nx<=Nxd; ++nx) {
      for (int nz=0; nz<=Nzd; ++nz)
        write(os, rdata_[flatten(nx,ny,nz,i)]);
    }
    for (int nx=Nx_-Nxd; nx<Nx_; ++nx) {
      for (int nz=0; nz<=Nzd; ++nz)
        write(os, rdata_[flatten(nx,ny,nz,i)]);
    }
      }
    }
  }
  else {
    int Ntotal = Nd_* Nx_ * Ny_ * Nzpad_;
    for (int i=0; i<Ntotal; ++i)
      write(os, rdata_[i]);
  }
}

void FlowField::b8Save(const string& filename, Real time) const {
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/

  FILE *fp;
  fp = fopen(filename.c_str(), "wb");
  if (fp == NULL) {
    cerr << "FlowField::b8Save : can't open file " << filename << endl;
    abort();
  }

  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();
  
  Real tmp[Nx_*Ny_*Nz_*Nd_+1];

  int i, j, k, d;
  
  tmp[0] = time;

  for (i=0; i<Nx_; ++i) {
    for (j=0; j<Ny_; ++j) {
        for (k=0; k<Nz_; ++k) {
            for (d=0; d<Nd_; ++d) {
                tmp[1+d+Nd_*(k+Nz_*(j+Ny_*i))] = u(i,j,k,d);
            }
        }
    }
  }

  fwrite(tmp, sizeof(Real), Nx_*Ny_*Nz_*Nd_+1, fp);
  fclose(fp);  

// recovery current state  
  u.makeState(xzstate, ystate);

}

void FlowField::b8Read(const string& filename, Real * time) const {
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/

  FILE *fp;
  fp = fopen(filename.c_str(), "rb");
  if (fp == NULL) {
    cerr << "FlowField::b8Read : can't open file " << filename << endl;
    abort();
  }

  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();
  
  Real tmp[Nx_*Ny_*Nz_*Nd_+1];

  fread(tmp, sizeof(Real), Nx_*Ny_*Nz_*Nd_+1, fp);
  fclose(fp);  

  int i, j, k, d;
  
  *time = tmp[0];

  for (i=0; i<Nx_; ++i) {
    for (j=0; j<Ny_; ++j) {
        for (k=0; k<Nz_; ++k) {
            for (d=0; d<Nd_; ++d) {
                u(i,j,k,d) = tmp[1+d+Nd_*(k+Nz_*(j+Ny_*i))];
            }
        }
    }
  }

//  if (Nd_ == 3) 
//  for (i=0; i<Nx_; ++i) {
//     for (k=0; k<Nz_; ++k) {
//             u(i,0,k,0) +=0.9;
//     }
//  }
  
// recovery current state  
  u.makeState(xzstate, ystate);
  u.makeSpectral();

}

Real FlowField::error_vs_laminarEkman(ChebyCoeff& Ubase, Real V, Real nu, Real omega) const {
  Real error;

  Real factor, D, f, alpha, tmpexp, tmpcos, tmpsin;
  f = 2.0*omega; D = pi*sqrt(2.0*nu/f); factor = pi/D;

  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;
// get real velocities 
  u.makePhysical();

  fieldstate Us = Ubase.state();
  Ubase.makePhysical();

  int i, j, k; Real eu, ev, ew;
  error = 0.0;
  for (j=0; j<Ny_; ++j) {
    alpha = factor*(b_-y(j)); tmpexp = exp(-alpha);
    tmpcos = V*cos(alpha)*tmpexp; tmpsin = V*sin(alpha)*tmpexp;
    for (i=0; i<Nx_; ++i) {
        for (k=0; k<Nz_; ++k) {
            eu = u(i,j,k,0)+Ubase[j]-tmpcos;
            ev = u(i,j,k,1);
            ew = u(i,j,k,2)-tmpsin;
            error += eu*eu+ev*ev+ew*ew;
        }
    }
  }
  error = sqrt(error/Real(Nx_*Ny_*Nz_));
  
// recovery current state  
  u.makeState(xzstate, ystate);
  Ubase.makeState(Us);
  
  return error;
}

Real FlowField::error_vs_laminarEkman(Real V, Real nu, Real omega) const {
  Real error;

  Real factor, D, f, alpha, tmpexp, tmpcos, tmpsin;
  f = 2.0*omega; D = pi*sqrt(2.0*nu/f); factor = pi/D;

  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;
// get real velocities 
  u.makePhysical();

  int i, j, k; Real eu, ev, ew;
  error = 0.0;
  for (j=0; j<Ny_; ++j) {
    alpha = factor*(b_-y(j)); tmpexp = exp(-alpha);
    tmpcos = V*cos(alpha)*tmpexp; tmpsin = V*sin(alpha)*tmpexp;
    for (i=0; i<Nx_; ++i) {
        for (k=0; k<Nz_; ++k) {
            eu = u(i,j,k,0)-tmpcos;
            ev = u(i,j,k,1);
            ew = u(i,j,k,2)-tmpsin;
            error += eu*eu+ev*ev+ew*ew;
        }
    }
  }
  error = sqrt(error/Real(Nx_*Ny_*Nz_));
  
// recovery current state  
  u.makeState(xzstate, ystate);
  
  return error;
}

void FlowField::plot3DQSave(const string& filebase, ChebyCoeff& Ubase, Real Re, Real Time) const {
  assert(Nd_==3);
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string filename = filebase;

  FILE *fp;
  fp = fopen(filename.c_str(), "wb");
  if (fp == NULL) {
    cerr << "FlowField::plot3DQ : can't open file " << filename << endl;
    abort();
  }

  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();

  int itmp[1];
  itmp[0]=Nx_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=Ny_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=Nz_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  
  int i, j, k;
  float tmp[4];

  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Re;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Time;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = 1.0;
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  fieldstate Us = Ubase.state();
  Ubase.makePhysical();
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = u(i,j,k,0)+Ubase[j];
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = u(i,j,k,1);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = u(i,j,k,2);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[1] = u(i,j,k,0)+Ubase[j];
            tmp[2] = u(i,j,k,1);
            tmp[3] = u(i,j,k,2);
            tmp[0] = (tmp[1]*tmp[1]+tmp[2]*tmp[2]+tmp[3]*tmp[3])*0.5;
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  Ubase.makeState(Us);

  fclose(fp);  

// recovery current state  
  u.makeState(xzstate, ystate);

}

void FlowField::plot3DQSave(const string& filebase, Real Re, Real Time) const {
  assert(Nd_==3);
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string filename = filebase;

  FILE *fp;
  fp = fopen(filename.c_str(), "wb");
  if (fp == NULL) {
    cerr << "FlowField::plot3DQ : can't open file " << filename << endl;
    abort();
  }

  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();

  int itmp[1];
  itmp[0]=Nx_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=Ny_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=Nz_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  
  int i, j, k;
  float tmp[4];

  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Re;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Time;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = 1.0;
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = u(i,j,k,0);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = u(i,j,k,1);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = u(i,j,k,2);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[1] = u(i,j,k,0);
            tmp[2] = u(i,j,k,1);
            tmp[3] = u(i,j,k,2);
            tmp[0] = (tmp[1]*tmp[1]+tmp[2]*tmp[2]+tmp[3]*tmp[3])*0.5;
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }

  fclose(fp);  

// recovery current state  
  u.makeState(xzstate, ystate);

}

void FlowField::plot3DXYZSave(const string& filebase) const {

  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string filename = filebase;

  FILE *fp;
  fp = fopen(filename.c_str(), "wb");
  if (fp == NULL) {
    cerr << "FlowField::plot3DXYZ : can't open file " << filename << endl;
    abort();
  }
  
  int itmp[1];
  itmp[0]=Nx_;
//  itmp[0] = SWAP(itmp[0]);
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=Ny_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=Nz_;
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  
  int i, j, k;
  float tmp[1];
 
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = x(i);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = y(j)-b_;
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[0] = z(k);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }

  fclose(fp);  
}

// save for tecplot 10
void FlowField::tecplot10Save(const std::string& filebase, Real time) const {
  assert(Nd_==3);

  static int zonenumber = 1;
  
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string fname = appendSuffix(filebase, ".plt");
  char filename[32];
  strcpy(filename, fname.c_str());
  
  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();

  int II,III,Debug,DIsDouble,VIsDouble,IMax,JMax,KMax;
  
  Debug = 0;
  DIsDouble = 0;
  VIsDouble = 0;
  IMax = Nx_;
  JMax = Ny_;
  KMax = Nz_;
  III = IMax*JMax*KMax;

  II = TECINI("Velocity Dataset","X Y Z U V W K",&filename[0],".",&Debug,&VIsDouble);

  II = TECZNE("Simple Zone",&IMax,&JMax,&KMax,"BLOCK",NULL);
  
  int i, j, k;
  float tmp[Nz_][Ny_][Nx_];
  float tmp1, tmp2, tmp3;

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = x(i); tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = y(j)-b_; tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = z(k); tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = u(i,j,k,0);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = u(i,j,k,1);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = u(i,j,k,2);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (i=0; i<Nx_; ++i) {
    for (j=0; j<Ny_; ++j) {
        for (k=0; k<Nz_; ++k) {
            tmp1 = u(i,j,k,0); tmp2 = u(i,j,k,1); tmp3 = u(i,j,k,2);
            tmp[k][j][i] = (tmp1*tmp1+tmp2*tmp2+tmp3*tmp3)*0.5;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  double xpos,ypos,fontheight,boxmargin,boxlinethickness,angle,linespacing;
  int poscoordmode,attachtozone,zone,font,fontheightunits;
  int boxtype,boxcolor,boxfillcolor,anchor,textcolor,scope;
  char text[32];

  xpos=70.;
  ypos=2.;
  poscoordmode=1;
  attachtozone=1;
  zone=zonenumber; //zonenumber++;
  font=1;
  fontheightunits=1;
  fontheight=3.;
  boxtype=0;
  boxmargin=5.;
  boxlinethickness=0.5;
  boxcolor=3;
  boxfillcolor=7;
  angle=0.;
  anchor=0;
  linespacing=1.5;
  textcolor=0;
  scope=1;

  sprintf(text, "Time=%10.3f", time);
  II = TECTXT(&xpos,&ypos,&poscoordmode,&attachtozone,&zone,&font,&fontheightunits,
        &fontheight,&boxtype,&boxmargin,&boxlinethickness,&boxcolor,&boxfillcolor,
        &angle,&anchor,&linespacing,&textcolor,&scope,&text[0],NULL);
  
  II = TECEND();
  
// recovery current state  
  u.makeState(xzstate, ystate);

}

// Open file for tecplot 10
void FlowField::tecplot10Open(const std::string& filebase) const {
  assert(Nd_==3);
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string fname = appendSuffix(filebase, ".plt");
  char filename[32];
  strcpy(filename, fname.c_str());
  
  int II,III,Debug,DIsDouble,VIsDouble,IMax,JMax,KMax;
  
  Debug = 0;
  DIsDouble = 0;
  VIsDouble = 0;
  IMax = Nx_;
  JMax = Ny_;
  KMax = Nz_;
  III = IMax*JMax*KMax;

  II = TECINI("Velocity Dataset","X Y Z U V W K",&filename[0],".",&Debug,&VIsDouble);
 
}

// save for tecplot 10
void FlowField::tecplot10Save(Real time) const {
  assert(Nd_==3);
  
  static int zonenumber = 1;
  
  FlowField& u = (FlowField&) *this;  // cast away constness
// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();

  int II,III,Debug,DIsDouble,VIsDouble,IMax,JMax,KMax;
  
  Debug = 0;
  DIsDouble = 0;
  VIsDouble = 0;
  IMax = Nx_;
  JMax = Ny_;
  KMax = Nz_;
  III = IMax*JMax*KMax;

  II = TECZNE("Simple Zone",&IMax,&JMax,&KMax,"BLOCK",NULL);
  
  int i, j, k;
  float tmp[Nz_][Ny_][Nx_];
  float tmp1, tmp2, tmp3;

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = x(i); tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = y(j)-b_; tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = z(k); tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = u(i,j,k,0);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = u(i,j,k,1);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = u(i,j,k,2);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (i=0; i<Nx_; ++i) {
    for (j=0; j<Ny_; ++j) {
        for (k=0; k<Nz_; ++k) {
            tmp1 = u(i,j,k,0); tmp2 = u(i,j,k,1); tmp3 = u(i,j,k,2);
            tmp[k][j][i] = (tmp1*tmp1+tmp2*tmp2+tmp3*tmp3)*0.5;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  double xpos,ypos,fontheight,boxmargin,boxlinethickness,angle,linespacing;
  int poscoordmode,attachtozone,zone,font,fontheightunits;
  int boxtype,boxcolor,boxfillcolor,anchor,textcolor,scope;
  char text[32];

  xpos=70.;
  ypos=2.;
  poscoordmode=1;
  attachtozone=1;
  zone=zonenumber;
  font=1;
  fontheightunits=1;
  fontheight=3.;
  boxtype=0;
  boxmargin=5.;
  boxlinethickness=0.5;
  boxcolor=3;
  boxfillcolor=7;
  angle=0.;
  anchor=0;
  linespacing=1.5;
  textcolor=0;
  scope=1;

  sprintf(text, "Time=%10.3f", time);
  II = TECTXT(&xpos,&ypos,&poscoordmode,&attachtozone,&zone,&font,&fontheightunits,
        &fontheight,&boxtype,&boxmargin,&boxlinethickness,&boxcolor,&boxfillcolor,
        &angle,&anchor,&linespacing,&textcolor,&scope,&text[0],NULL);

  zonenumber++;
  
// recovery current state  
  u.makeState(xzstate, ystate);

}

// Close file for tecplot 10
void FlowField::tecplot10Close() const {TECEND();}

// save U profile using wall coordinates
void FlowField::saveU(const std::string& filebase, Real nu) const
{
  FlowField& u = (FlowField&) *this;  // cast away constness
  assert(u.Nd() == 3);

// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();

  FlowField U(1,Ny_,1,1,Lx_,Lz_,a_,b_);
  FlowField dUdy(1,Ny_,1,1,Lx_,Lz_,a_,b_);

  U.makePhysical();
  int i,j,k;
  for (j=0; j<Ny_; ++j) {
    U(0,j,0,0) = 0.0;
    for (i=0; i<Nx_; ++i) {
        for (k=0; k<Nz_; ++k) {
        U(0,j,0,0) += u(i,j,k,0);
        }
    }
    U(0,j,0,0) = U(0,j,0,0)/((double)Nx_*Nz_);
  }
// recovery current state  
  u.makeState(xzstate, ystate);

  ydiff(U, dUdy);
  dUdy.makePhysical();
  
  Real uu, yy;
  uu = dUdy(0,Ny_-1,0,0)-dUdy(0,0,0,0);
  uu = sqrt(nu*fabs(uu)*0.5);
  
  string filename = filebase;
  FILE *fpu;
  fpu = fopen(filename.c_str(), "w");
  if (fpu == NULL) {
    cerr << "FlowField::saveU can't open file " << endl;
    abort();
  }

  fprintf(fpu, "#y u\n");
  for (j=0; j<=(Ny_-1)/2; ++j) {
    yy = (b_-y(j))*uu/nu;
    fprintf(fpu, "%15.7e %15.7e\n", yy, (U(0,j,0,0)+U(0,Ny_-1-j,0,0))*0.5/(uu+1.0e-30));
  }
  fclose(fpu);  
  
}

// save U defect profile using wall coordinates, velocity at bottom is not defect
void FlowField::saveUd(const std::string& filebase, Real nu, Real Usurf) const
{
  FlowField& u = (FlowField&) *this;  // cast away constness
  assert(u.Nd() == 3);

// backup current state
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;

// get real velocities 
  u.makePhysical();

  FlowField U(1,Ny_,1,1,Lx_,Lz_,a_,b_);
  FlowField dUdy(1,Ny_,1,1,Lx_,Lz_,a_,b_);

  FlowField W(1,Ny_,1,1,Lx_,Lz_,a_,b_);
  FlowField dWdy(1,Ny_,1,1,Lx_,Lz_,a_,b_);

  U.makePhysical();
  W.makePhysical();
  int i,j,k;
  for (j=0; j<Ny_; ++j) {
    U(0,j,0,0) = 0.0;
    for (i=0; i<Nx_; ++i) {
        for (k=0; k<Nz_; ++k) {
        U(0,j,0,0) += u(i,j,k,0);
        W(0,j,0,0) += u(i,j,k,2);
        }
    }
    U(0,j,0,0) = U(0,j,0,0)/((double)Nx_*Nz_);
    W(0,j,0,0) = W(0,j,0,0)/((double)Nx_*Nz_);
    //U(0,j,0,0) = Usurf-U(0,j,0,0);
  }
// recovery current state  
  u.makeState(xzstate, ystate);

  ydiff(U, dUdy);
  dUdy.makePhysical();
  ydiff(W, dWdy);
  dWdy.makePhysical();
  
  Real uu, yy, uu_b, yy_b;
  uu = dUdy(0,0,0,0);
  uu = sqrt(nu*fabs(uu));
  uu_b = dUdy(0,Ny_-1,0,0);
  uu_b = sqrt(nu*fabs(uu_b));

  Real ww, yyw, ww_b, yyw_b;
  ww = dWdy(0,0,0,0);
  ww = sqrt(nu*fabs(ww));
  ww_b = dWdy(0,Ny_-1,0,0);
  ww_b = sqrt(nu*fabs(ww_b));
  
  string filename = filebase;
  FILE *fpu;
  fpu = fopen(filename.c_str(), "w");
  if (fpu == NULL) {
    cerr << "FlowField::saveUd can't open file " << endl;
    abort();
  }

  fprintf(fpu, "#y u yb ub yw w ywb wb\n");
  for (j=0; j<=(Ny_-1)/2; ++j) {
    yy = fabs((b_-y(j))*uu/nu);
    yy_b = fabs((y(Ny_-1-j)-a_)*uu_b/nu);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e ",
       yy, fabs(Usurf-U(0,j,0,0))/(uu+1.0e-40), yy_b, fabs(U(0,Ny_-1-j,0,0)/(uu_b+1.0e-40)));
    yyw = fabs((b_-y(j))*ww/nu);
    yyw_b = fabs((y(Ny_-1-j)-a_)*ww_b/nu);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e\n",
       yyw, fabs(W(0,j,0,0))/(ww+1.0e-40), yyw_b, fabs(W(0,Ny_-1-j,0,0)/(ww_b+1.0e-40)));
  }
  fclose(fpu);  
  
}

// save evolution
void FlowField::evolution_ini() {

  k_ = 0.0;
  
  FILE *fp;
  fp = fopen("evolution.dat", "w");
  if (fp == NULL) {
    cerr << "FlowField::evolution_ini can't open file " << endl;
    abort();
  }

  fprintf(fp, "#time step KE TKE PE ReTauTop ReTauBot ReLambda ReLambdaMax "); //9
  fprintf(fp, "Fr2 Fr2Max Fr2Min Ri RiMax RiMin RoW2 RoW2Max "); //8(+9=17)
  fprintf(fp, "uu vv ww uv wv uw ss us vs ws "); //10(+17=27)
  fprintf(fp, "fuu fvv fww ftke fus fvs fws fss fpe "); //9(+27=36)
  fprintf(fp, "rruu rrww rbtke rrus rbvs rrws rbpe "); //7(+36=43)
  fprintf(fp, "KEc KE01 KE19 PEc PE01 PE19 MKEc MKE01 MKE19 MPEc MPE01 MPE19 "); //12(+43=55)
  fprintf(fp, "DIS DIV RaMax Ycgc Ycg01 Ycg19 dsdyc dsdy01 dsdy19\n"); //9(+55=64)
  
  fclose(fp);  

}

void FlowField::evolution(Real time, int stp, Real *Measures) {

  k_ = Measures[0];
  
  FILE *fp;
  fp = fopen("evolution.dat", "a");
  if (fp == NULL) {
    cerr << "FlowField::evolution can't open file " << endl;
    abort();
  }
  
  fprintf(fp, "%15.7e %08d ", time, stp);
  for(int i=0; i<62; ++i) fprintf(fp, "%15.7e ", Measures[i]);
  fprintf(fp, "\n");
    
  fclose(fp);

}

void FlowField::dump() const {
  for (int i=0; i<Nx_ * Ny_ * Nzpad_ * Nd_; ++i)
    cout << rdata_[i] << ' ';
  cout << endl;
}

Real FlowField::energy(bool normalize) const {
  assert(xzstate_ == Spectral && ystate_ == Spectral);
  return L2Norm2(*this, normalize);
}
Real FlowField::energy(int mx, int mz, bool normalize) const {
  assert(xzstate_ == Spectral && ystate_ == Spectral);
  ComplexChebyCoeff u(Ny_,a_,b_,Spectral);
  Real e = 0.0;
  for (int i=0; i<Nd_; ++i) {
    for (int ny=0; ny<Ny_; ++ny) 
      u.set(ny,this->cmplx(mx,ny,mz,i));
    e += L2Norm2(u, normalize);
  }
  if (!normalize)
    e *= Lx_*Lz_;
  return e;
}
Real FlowField::energy(int mx, int mz, int iy, bool normalize) const {
  assert(xzstate_ == Spectral && ystate_ == Spectral);
  ComplexChebyCoeff u(1,a_,b_,Spectral);
  Real e = 0.0;
  for (int i=0; i<Nd_; ++i) {
    u.set(0,this->cmplx(mx,iy,mz,i));
    e += L2Norm2(u, normalize);
  }
  if (!normalize)
    e *= Lx_*Lz_;
  return e;
}

Real FlowField::dudy_a() const {
  assert(ystate_ == Spectral);
  BasisFunc prof = profile(0,0);
  ChebyCoeff dudy = diff(Re(prof.u()));
  return dudy.eval_a();
}
Real FlowField::dudy_b() const {
  assert(ystate_ == Spectral);
  BasisFunc prof = profile(0,0);
  ChebyCoeff dudy = diff(Re(prof.u()));
  return dudy.eval_b();
}

Real FlowField::CFLfactor() const {
  assert(Nd_ == 3);
  FlowField& u = (FlowField&) *this;
  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;
  u.makePhysical();
  Vector y = chebypoints(Ny_, a_, b_);
  Real cfl = 0.0;
  Real magu;
  Vector dx(3);
  dx[0] = Lx_/Nx_;
  dx[2] = Lz_/Nz_;
  for (int ny=0; ny<Ny_; ++ny) { 
      dx[1] = (ny==0 || ny==Ny_-1) ? (y[0]-y[2])*0.5 : (y[ny-1]-y[ny+1])*0.5;
      for (int nx=0; nx<Nx_; ++nx) {
      for (int nz=0; nz<Nz_; ++nz) {
        magu = sqrt(u(nx,ny,nz,0)*u(nx,ny,nz,0)+u(nx,ny,nz,1)*u(nx,ny,nz,1)+u(nx,ny,nz,2)*u(nx,ny,nz,2));
        for (int i=0; i<Nd_; ++i) cfl = Greater(cfl, magu/dx[i]);
      }
      }
  }
  u.makeState(xzstate,ystate);

  return cfl;
}

Real FlowField::CFLfactor(const ChebyCoeff& Ubase_) const {
  if (Ubase_.N() == 0)
    return CFLfactor();

  assert(Nd_ == 3);
  FlowField& u = const_cast<FlowField&>(*this);
  ChebyCoeff& Ubase = const_cast<ChebyCoeff&>(Ubase_);

  fieldstate xzstate = xzstate_;
  fieldstate ystate = ystate_;
  fieldstate Ustate = Ubase.state();
  u.makePhysical();
  Ubase.makePhysical();
  Vector y = chebypoints(Ny_, a_, b_);
  Real cfl = 0.0;
  Real magu;
  Vector dx(3);
  dx[0] = Lx_/Nx_;
  dx[2] = Lz_/Nz_;
  for (int ny=0; ny<Ny_; ++ny) {
      dx[1] = (ny==0 || ny==Ny_-1) ? (y[0]-y[2])*0.5 : (y[ny-1]-y[ny+1])*0.5;
      for (int nx=0; nx<Nx_; ++nx) {
      for (int nz=0; nz<Nz_; ++nz) {
        magu = sqrt((u(nx,ny,nz,0)+Ubase(ny))*(u(nx,ny,nz,0)+Ubase(ny))+u(nx,ny,nz,1)*u(nx,ny,nz,1)+u(nx,ny,nz,2)*u(nx,ny,nz,2));
        for (int i=0; i<3; ++i) cfl = Greater(cfl, magu/dx[i]);
      }
      }
    }
  u.makeState(xzstate,ystate);
  Ubase.makeState(Ustate);

  return cfl;
}


/********************************************************
Real FlowField::dissipation(int mx, int mz, Real nu) const {
  assert(xzstate_ == Spectral && ystate_== Spectral);
  BasisFunc prof;
  Real d=0.0;
  for (int _=0; _<Nd_; ++_) {
    prof = profile(mx, mz);
    d += nu*Re(L2InnerProduct(prof, ::laplacian(prof), false));
  }
  return d;
}
***********************************************************/

void FlowField::setState(fieldstate xz, fieldstate y) {
  xzstate_ = xz;
  ystate_ = y;
}
void FlowField::assertState(fieldstate xz, fieldstate y) const {
  assert(xzstate_ == xz && ystate_ == y);
}

void swap(FlowField& f, FlowField& g) {
  assert(f.congruent(g));
  Real* tmp = f.rdata_;
  f.rdata_ = g.rdata_;
  g.rdata_ = tmp;

  f.cdata_ = (Complex*)f.rdata_;
  g.cdata_ = (Complex*)g.rdata_;

  // We must swap the FFTW3 plans, as well, since they're tied to specific 
  // memory locations. This burned me when I upgraded FFTW2->FFTW3.
  fftw_plan tmp_plan;
  tmp_plan = f.xz_plan_;
  f.xz_plan_ = g.xz_plan_;
  g.xz_plan_ = tmp_plan;

  tmp_plan = f.xz_iplan_;
  f.xz_iplan_ = g.xz_iplan_;
  g.xz_iplan_ = tmp_plan;
  
  tmp_plan = f.y_plan_;
  f.y_plan_ = g.y_plan_;
  g.y_plan_ = tmp_plan;
}
  
void orthonormalize(array<FlowField>& e) {

  // Modified Gram-Schmidt orthogonalization
  int N = e.length();
  FlowField em_tmp;
  
  for (int m=0; m<N; ++m) {
    FlowField& em = e[m];
    em *= 1.0/L2Norm(em);
    em *= 1.0/L2Norm(em); // reduces floating-point errors

    // orthogonalize
    for (int n=m+1; n<N; ++n) {
      em_tmp = em;
      em_tmp *= L2InnerProduct(em, e[n]);
      e[n] -= em_tmp;
    }
  }
  return;
}

void normalize(array<FlowField>& e) {
  int N = e.length();
  for (int m=0; m<N; ++m) {
    FlowField& em = e[m];
    em *= 1.0/L2Norm(em);
    em *= 1.0/L2Norm(em);
  }
  return;
}

void orthogonalize(array<FlowField>& e) {
  // Modified Gram-Schmidt orthogonalization
  int N = e.length();
  FlowField em_tmp;
  
  for (int m=0; m<N; ++m) {
    FlowField& em = e[m];
    Real emnorm2 = L2Norm2(em);
    for (int n=m+1; n<N; ++n) {
      em_tmp = em;
      em_tmp *= L2InnerProduct(em, e[n])/emnorm2;
      e[n] -= em_tmp;
    }
  }
  return;
}

void mean_profiles(const std::string& filebase, FlowField& u, FlowField& q, FlowField& s, 
     Real nu, Real kappa, Real g, Real Omega, Real UdL,
     Real *Measures, FlowField& LRi, FlowField& MProfile) {
  assert(u.Nd()==3); assert(q.Nd()==1); assert(s.Nd()==1); 
  assert(LRi.Nd()==1); assert(MProfile.Nd()==4);

// backup current state
  fieldstate uxzstate = u.xzstate();
  fieldstate uystate = u.ystate();
  fieldstate qxzstate = q.xzstate();
  fieldstate qystate = q.ystate();
  fieldstate sxzstate = s.xzstate();
  fieldstate systate = s.ystate();

  int ix, iy, iz;
  int Nx, Ny, Nz;
  Nx = u.Nx(); Ny=u.Ny(); Nz=u.Nz();
  Real tmp1, tmp2, tmp3;

  Real dis_[Ny];
  
  Real div_[Ny];

  FlowField gradu(Nx,Ny,Nz,3,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField dsdy(Nx,Ny,Nz,1,s.Lx(),s.Lz(),s.a(),s.b());

  LRi.makePhysical();
  
  grad(u[0], gradu);
  u.makePhysical();
  gradu.makePhysical();
  for (iy=0; iy<Ny; ++iy) {
    dis_[iy] = 0.0;
    div_[iy] = 0.0;
    for (ix=0; ix<Nx; ++ix) {
        for (iz=0; iz<Nz; ++iz) {
            dis_[iy] +=square(gradu(ix,iy,iz,0));
            dis_[iy] +=square(gradu(ix,iy,iz,1));
            dis_[iy] +=square(gradu(ix,iy,iz,2));
            div_[iy] +=gradu(ix,iy,iz,0);
        }
    }
    LRi(0,iy,0,0) = u(Nx/2,iy,Nz/2,0)*gradu(Nx/2,iy,Nz/2,1);
  }
  
  grad(u[2], gradu);
  u.makePhysical();
  gradu.makePhysical();
  for (iy=0; iy<Ny; ++iy) {
    for (ix=0; ix<Nx; ++ix) {
        for (iz=0; iz<Nz; ++iz) {
            dis_[iy] +=square(gradu(ix,iy,iz,0));
            dis_[iy] +=square(gradu(ix,iy,iz,1));
            dis_[iy] +=square(gradu(ix,iy,iz,2));
            div_[iy] +=gradu(ix,iy,iz,2);
        }
    }
    tmp1 = LRi(0,iy,0,0)+u(Nx/2,iy,Nz/2,2)*gradu(Nx/2,iy,Nz/2,1);
    LRi(0,iy,0,0) = (square(u(Nx/2,iy,Nz/2,0))+square(u(Nx/2,iy,Nz/2,2)))/(square(tmp1)+1.0e-30);
  }

  grad(u[1], gradu);
  gradu.makePhysical();
  for (iy=0; iy<Ny; ++iy) {
    for (ix=0; ix<Nx; ++ix) {
        for (iz=0; iz<Nz; ++iz) {
            dis_[iy] +=square(gradu(ix,iy,iz,0));
            dis_[iy] +=square(gradu(ix,iy,iz,1));
            dis_[iy] +=square(gradu(ix,iy,iz,2));
            div_[iy] +=gradu(ix,iy,iz,1);
        }
    }
    dis_[iy] = dis_[iy]*nu/((double)Nx*Nz);
    div_[iy] = div_[iy]/((double)Nx*Nz);
  }

  ydiff(s, dsdy);
  dsdy.makePhysical();
  for (iy=0; iy<Ny; ++iy) {
    LRi(0,iy,0,0) = -g*LRi(0,iy,0,0)*dsdy(Nx/2,iy,Nz/2,0);
  }
  
  Real Ycg_[Ny];
  for (iy=0; iy<Ny; ++iy) {
    Ycg_[iy] = 0.0;
    for (ix=0; ix<Nx; ++ix) {
        for (iz=0; iz<Nz; ++iz) {
        if (dsdy(ix,iy,iz,0) > 0.0) Ycg_[iy]++;
        }
    }
    Ycg_[iy] = 100.0*Ycg_[iy]/((double)Nx*Nz);
  }
  
  //Real Dx, Dz, Dxz, Delta;
  //Dx = u.Lx()/Nx; Dz = u.Lz()/Nz; Dxz = (Dx > Dz)? Dx:Dz;
  Real Delta;
  tmp1 = -1.e8; tmp2 = g/(nu*kappa);
  for (iy=0; iy<Ny; ++iy) {
    Delta = (iy == Ny-1)? fabs(u.y(iy-1)-u.y(iy)):fabs(u.y(iy)-u.y(iy+1));
    //Delta = (Delta > Dxz)? Delta:Dxz;
    Delta = tmp2*square(square(Delta));
    for (ix=0; ix<Nx; ++ix) {
        for (iz=0; iz<Nz; ++iz) {
           tmp3 = dsdy(ix,iy,iz,0)*Delta;
           tmp1 = (tmp3 > tmp1)? tmp3:tmp1;
        }
    }
  }
  Measures[55] = tmp1;
  
  Real m_dis = 0.0;
  Real m_div = 0.0;
  for (iy=1; iy<Ny; ++iy) {
     m_dis +=(dis_[iy-1]+dis_[iy])*(u.y(iy-1)-u.y(iy));
     m_div +=(div_[iy-1]+div_[iy])*(u.y(iy-1)-u.y(iy));
  }
  m_dis = m_dis*0.5;
  m_div = m_div*0.5;
  Measures[53] = m_dis;
  Measures[54] = m_div;

  Real w2_[Ny];
  
  curl(u, gradu);
  gradu.makePhysical();
  for (iy=0; iy<Ny; ++iy) {
    w2_[iy] = 0.0;
    for (ix=0; ix<Nx; ++ix) {
        for (iz=0; iz<Nz; ++iz) {
            w2_[iy] +=square(gradu(ix,iy,iz,1));
        }
    }
    w2_[iy] = sqrt(w2_[iy]/((double)Nx*Nz));
  }
  
  Real m_w2 = 0.0, w2_max=w2_[0];
  for (iy=1; iy<Ny; ++iy) {
  m_w2 += (w2_[iy-1]+w2_[iy])*(u.y(iy-1)-u.y(iy));
  w2_max = (w2_max > w2_[iy])? w2_max:w2_[iy];
  }
  m_w2 = m_w2*0.5/(u.b()-u.a());

  u.makePhysical();
  q.makePhysical();
  s.makePhysical();
  
  FlowField u_(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField v_(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField w_(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField q_(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField s_(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());
  u_.makePhysical();
  v_.makePhysical();
  w_.makePhysical();
  q_.makePhysical();

  Real uu_[Ny], vv_[Ny], ww_[Ny], uv_[Ny], wv_[Ny], uw_[Ny];
  Real Ri_[Ny], Fr2_[Ny], Row2_[Ny], ReT_[Ny], alpha_[Ny];
  Real ss_[Ny], vs_[Ny], us_[Ny], ws_[Ny], pe_[Ny], ke_[Ny], tke_[Ny];
  FlowField du_dy(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField dw_dy(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField ds_dy(1,u.Ny(),1,1,u.Lx(),u.Lz(),u.a(),u.b());

  s_.makePhysical();

  MProfile.makePhysical();
  for(iy = 0; iy<Ny; iy++) {
    u_(0,iy,0,0) = 0.0; v_(0,iy,0,0) = 0.0; w_(0,iy,0,0) = 0.0;
    q_(0,iy,0,0) = 0.0; s_(0,iy,0,0) = 0.0;
    for(ix = 0; ix<Nx; ix++) {
      for(iz = 0; iz<Nz; iz++) {
      u_(0,iy,0,0) += u(ix,iy,iz,0);
      v_(0,iy,0,0) += u(ix,iy,iz,1);
      w_(0,iy,0,0) += u(ix,iy,iz,2);
      q_(0,iy,0,0) += q(ix,iy,iz,0);
      s_(0,iy,0,0) += s(ix,iy,iz,0);
      }
    }
    tmp1 = 1.0/((double)Nx*Nz);
    u_(0,iy,0,0) = u_(0,iy,0,0)*tmp1;
    v_(0,iy,0,0) = v_(0,iy,0,0)*tmp1;
    w_(0,iy,0,0) = w_(0,iy,0,0)*tmp1;
    q_(0,iy,0,0) = q_(0,iy,0,0)*tmp1;
    s_(0,iy,0,0) = s_(0,iy,0,0)*tmp1;
    MProfile(0,iy,0,0) = u_(0,iy,0,0);
    MProfile(0,iy,0,1) = v_(0,iy,0,0);
    MProfile(0,iy,0,2) = w_(0,iy,0,0);
    MProfile(0,iy,0,3) = s_(0,iy,0,0);
    pe_[iy] = s_(0,iy,0,0)*g*(u.y(iy)-u.a());

    uu_[iy] = 0.0; vv_[iy] = 0.0; ww_[iy] = 0.0; 
    uv_[iy] = 0.0; uw_[iy] = 0.0; wv_[iy] = 0.0;
    ss_[iy] = 0.0; ke_[iy] = 0.0; 
    vs_[iy] = 0.0; us_[iy] = 0.0; ws_[iy] = 0.0;
    for(ix = 0; ix<Nx; ix++) {
      for(iz = 0; iz<Nz; iz++) {
      uu_[iy] += square(u(ix,iy,iz,0)-u_(0,iy,0,0));
      vv_[iy] += square(u(ix,iy,iz,1)-v_(0,iy,0,0));
      ww_[iy] += square(u(ix,iy,iz,2)-w_(0,iy,0,0));
      ss_[iy] += square(s(ix,iy,iz,0)-s_(0,iy,0,0));
      ke_[iy] += square(u(ix,iy,iz,0))+square(u(ix,iy,iz,1))+square(u(ix,iy,iz,2));

      uv_[iy] += (u(ix,iy,iz,0)-u_(0,iy,0,0))*(u(ix,iy,iz,1)-v_(0,iy,0,0));
      uw_[iy] += (u(ix,iy,iz,0)-u_(0,iy,0,0))*(u(ix,iy,iz,2)-w_(0,iy,0,0));
      wv_[iy] += (u(ix,iy,iz,2)-w_(0,iy,0,0))*(u(ix,iy,iz,1)-v_(0,iy,0,0));

      us_[iy] += (u(ix,iy,iz,0)-u_(0,iy,0,0))*(s(ix,iy,iz,0)-s_(0,iy,0,0));
      vs_[iy] += (u(ix,iy,iz,1)-v_(0,iy,0,0))*(s(ix,iy,iz,0)-s_(0,iy,0,0));
      ws_[iy] += (u(ix,iy,iz,2)-w_(0,iy,0,0))*(s(ix,iy,iz,0)-s_(0,iy,0,0));
      }
    }
    uu_[iy] = uu_[iy]*tmp1;
    vv_[iy] = vv_[iy]*tmp1;
    ww_[iy] = ww_[iy]*tmp1;
    uv_[iy] = uv_[iy]*tmp1;
    uw_[iy] = uw_[iy]*tmp1;
    wv_[iy] = wv_[iy]*tmp1;
    ss_[iy] = ss_[iy]*tmp1;
    ke_[iy] = 0.5*ke_[iy]*tmp1;
    tke_[iy] = 0.5*(uu_[iy]+vv_[iy]+ww_[iy]);
    vs_[iy] = vs_[iy]*tmp1;
    us_[iy] = us_[iy]*tmp1;
    ws_[iy] = ws_[iy]*tmp1;
    Row2_[iy] = w2_[iy]*0.5/(Omega+1.0e-30);
    ReT_[iy] = sqrt(20.0/3.0/(dis_[iy]*nu+1.0e-30))*tke_[iy];
    
  }

  ix = (Nx-1)/2; iy = (Ny-1)/2; iz = (Nz-1)/2;
  Measures[41] = 0.5*(square(u(ix,iy,iz,0)-u_(0,iy,0,0))+square(u(ix,iy,iz,1)-v_(0,iy,0,0))+square(u(ix,iy,iz,2)-w_(0,iy,0,0)));
  Measures[47] = ke_[iy];
  const Real location = 0.1; //two locations will be read: y=-location and y=-(Ly-location)
  static int idy = 0;
  if (idy == 0) {
     tmp3 = 2.0/(u.b()-u.a());
     iy = 0;
     do {
       iy++;
       tmp1 = (u.b()-u.y(iy))*tmp3;
     } while (tmp1 < location);
     idy = iy;
  }
  iy = idy;
  Measures[42] = 0.5*(square(u(ix,iy,iz,0)-u_(0,iy,0,0))+square(u(ix,iy,iz,1)-v_(0,iy,0,0))+square(u(ix,iy,iz,2)-w_(0,iy,0,0)));;
  Measures[48] = ke_[iy];
  iy = Ny-idy;
  Measures[43] = 0.5*(square(u(ix,iy,iz,0)-u_(0,iy,0,0))+square(u(ix,iy,iz,1)-v_(0,iy,0,0))+square(u(ix,iy,iz,2)-w_(0,iy,0,0)));;
  Measures[49] = ke_[iy];

  iy = (Ny-1)/2;
  Measures[44] = s(ix,iy,iz,0)*g*(u.y(iy)-u.a());
  Measures[50] = pe_[iy];
  iy = idy;
  Measures[45] = s(ix,iy,iz,0)*g*(u.y(iy)-u.a());
  Measures[51] = pe_[iy];
  iy = Ny-idy;
  Measures[46] = s(ix,iy,iz,0)*g*(u.y(iy)-u.a());
  Measures[52] = pe_[iy];

  iy = (Ny-1)/2;
  Measures[56] = Ycg_[iy];
  Measures[59] = dsdy(Nx/2,iy,Nz/2,0);
  iy = idy;
  Measures[57] = Ycg_[iy];
  Measures[60] = dsdy(Nx/2,iy,Nz/2,0);
  iy = Ny-idy;
  Measures[58] = Ycg_[iy];
  Measures[61] = dsdy(Nx/2,iy,Nz/2,0);
  
  //alpha_[0] = 0.0;
  for(iy = 0; iy<Ny-1; iy++) {
    tmp1 = w_(0,iy,0,0);
    tmp2 = u_(0,iy,0,0);
    tmp3 = (abs(tmp2) < 1.0e-5)? 0.0:acos(tmp2/(sqrt(tmp1*tmp1+tmp2*tmp2)+1.0e-30));
    alpha_[iy] = (tmp1 >=0.0)? tmp3:2.0*pi-tmp3;
  }
  alpha_[Ny-1] = alpha_[Ny-2];
  
  ydiff(u_, du_dy);
  ydiff(w_, dw_dy);
  ydiff(s_, ds_dy);
  ds_dy.makePhysical();
  du_dy.makePhysical();
  dw_dy.makePhysical();

  Real fuu_[Ny], fvv_[Ny], fww_[Ny], ftke_[Ny];
  Real fus_[Ny], fvs_[Ny], fws_[Ny], fss_[Ny], fpe_[Ny];
  Real rruu_[Ny], rrww_[Ny], rbtke_[Ny];
  Real rrus_[Ny], rbvs_[Ny], rrws_[Ny], rbpe_[Ny];
  Real uvdudy_[Ny], wvdwdy_[Ny];
  Real uvdsdy_[Ny], vvdsdy_[Ny], wvdsdy_[Ny], vsdudy_[Ny], vsdwdy_[Ny];
  for(iy = 0; iy<Ny; iy++) {
    fuu_[iy] = -4.0*Omega*uw_[iy]-2.0*uv_[iy]*du_dy(0,iy,0,0);
    fvv_[iy] = -2.0*g*vs_[iy];
    fww_[iy] = 4.0*Omega*uw_[iy]-2.0*wv_[iy]*dw_dy(0,iy,0,0);
    ftke_[iy] = -g*vs_[iy]-uv_[iy]*du_dy(0,iy,0,0)-wv_[iy]*dw_dy(0,iy,0,0);
    fus_[iy] = -2.0*Omega*ws_[iy]-uv_[iy]*ds_dy(0,iy,0,0)-vs_[iy]*du_dy(0,iy,0,0);
    fvs_[iy] = -g*ss_[iy]-vv_[iy]*ds_dy(0,iy,0,0);
    fws_[iy] = 2.0*Omega*us_[iy]-wv_[iy]*ds_dy(0,iy,0,0)-vs_[iy]*dw_dy(0,iy,0,0);
    fss_[iy] = -2.0*vs_[iy]*ds_dy(0,iy,0,0);
    fpe_[iy] = -2*g*kappa*ds_dy(0,iy,0,0)+g*vs_[iy];

    rruu_[iy] = -4.0*Omega*uw_[iy]/(-2.0*uv_[iy]*du_dy(0,iy,0,0)+1.0e-30);
    rrww_[iy] = 4.0*Omega*uw_[iy]/(-2.0*wv_[iy]*dw_dy(0,iy,0,0)+1.0e-30);
    rbtke_[iy] = -g*vs_[iy]/(-uv_[iy]*du_dy(0,iy,0,0)-wv_[iy]*dw_dy(0,iy,0,0)+1.0e-30);
    rrus_[iy] = -2.0*Omega*ws_[iy]/(-uv_[iy]*ds_dy(0,iy,0,0)-vs_[iy]*du_dy(0,iy,0,0)+1.0e-30);
    rbvs_[iy] = -g*ss_[iy]/(-vv_[iy]*ds_dy(0,iy,0,0)+1.0e-30);
    rrws_[iy] = 2.0*Omega*us_[iy]/(-wv_[iy]*ds_dy(0,iy,0,0)-vs_[iy]*dw_dy(0,iy,0,0)+1.0e-30);
    rbpe_[iy] = g*vs_[iy]/(-2*g*kappa*ds_dy(0,iy,0,0)+1.0e-30);
    
    uvdudy_[iy] = uv_[iy]*du_dy(0,iy,0,0);
    wvdwdy_[iy] = wv_[iy]*dw_dy(0,iy,0,0);
    uvdsdy_[iy] = uv_[iy]*ds_dy(0,iy,0,0);
    vvdsdy_[iy] = vv_[iy]*ds_dy(0,iy,0,0);
    wvdsdy_[iy] = wv_[iy]*ds_dy(0,iy,0,0);
    vsdudy_[iy] = vs_[iy]*du_dy(0,iy,0,0);
    vsdwdy_[iy] = vs_[iy]*dw_dy(0,iy,0,0);
  }
  
  //Real factor=-square(2.0*pi*UdL)/g;
  Real factor=-square(UdL)/g; 
  Real Nmax = 0.0; 
  Real N_[Ny], krho_[Ny], norkrho_[Ny], nordis_[Ny];
  for(iy = 0; iy<Ny; iy++) {
    tmp1 = ds_dy(0,iy,0,0);
    tmp2 = -g*tmp1;
    tmp3 = square(u_(0,iy,0,0)*du_dy(0,iy,0,0)+w_(0,iy,0,0)*dw_dy(0,iy,0,0));
    Ri_[iy] = tmp2*(square(u_(0,iy,0,0))+square(w_(0,iy,0,0)))/(tmp3+1.0e-30);
    Fr2_[iy] = factor/(tmp1+1.0e-30);
    N_[iy] = (tmp2 > 0.0) ? sqrt(tmp2):0.0;
    Nmax = (N_[iy] > Nmax)? N_[iy]:Nmax;
    krho_[iy] = g*vs_[iy]/(tmp2+1.0e-30);
    norkrho_[iy] = krho_[iy]/kappa;
    nordis_[iy] = dis_[iy]/(nu*(tmp2+1.0e-30));
  }
  s.setNmax(Nmax);

  Real dy;
  Real Rimin=Ri_[0];
  Real Rimax=Ri_[0];
  Real Fr2min=Fr2_[0];
  Real Fr2max=Fr2_[0];
  Real ReLambdaMax=ReT_[0];
  Real Row2Max=Row2_[0];
  Real i_uu, i_vv, i_ww, i_uv, i_wv, i_uw;
  Real i_ss, i_us, i_vs, i_ws;
  Real m_ri, m_fr2, m_reT, m_row2, i_ke, i_tke, i_pe;
  Real i_fuu, i_fvv, i_fww, i_ftke, i_fus, i_fvs, i_fws, i_fss, i_fpe;
//  Real i_rruu, i_rrww, i_rbtke, i_rrus, i_rbvs, i_rrws, i_rbpe;
  Real i_uvdudy, i_wvdwdy, i_uvdsdy, i_vvdsdy, i_wvdsdy, i_vsdudy, i_vsdwdy, i_dsdy;
  i_uu=0.;i_vv=0.;i_ww=0.;i_uv=0.;i_wv=0.;i_uw=0.;
  i_ss=0.;i_us=0.;i_vs=0.;i_ws=0.;m_ri=0.;m_fr2=0.;
  m_reT=0.;m_row2=0.;i_ke=0.;i_tke=0.;i_pe=0.;
  i_fuu=0.;i_fvv=0.;i_fww=0.;i_ftke=0.;i_fus=0.;i_fvs=0.;i_fws=0.;i_fss=0.;i_fpe=0.;
//  i_rruu=0.;i_rrww=0.;i_rbtke=0.;i_rrus=0.;i_rbvs=0.;i_rrws=0.;i_rbpe=0.;
  i_uvdudy=0.;i_wvdwdy=0.;i_uvdsdy=0.;i_vvdsdy=0.;i_wvdsdy=0.;i_vsdudy=0.;i_vsdwdy=0.;i_dsdy=0.;
  for(iy = 1; iy<Ny; iy++) {
    dy = u.y(iy-1)-u.y(iy);

    Rimin = (Ri_[iy] > 0.0 && Ri_[iy] < Rimin) ? Ri_[iy]:Rimin;
    Rimax = (Ri_[iy] > Rimax) ? Ri_[iy]:Rimax;
    tmp1 = (Ri_[iy-1] > 0.0)? Ri_[iy-1]:0.0;
    tmp2 = (Ri_[iy] > 0.0)? Ri_[iy]:0.0;
    m_ri +=(tmp1+tmp2)*dy;

    Fr2min = (Fr2_[iy] > 0.0 && Fr2_[iy] < Fr2min) ? Fr2_[iy]:Fr2min;
    Fr2max = (Fr2_[iy] > Fr2max) ? Fr2_[iy]:Fr2max;
    tmp1 = (Fr2_[iy-1] > 0.0)? Fr2_[iy-1]:0.0;
    tmp2 = (Fr2_[iy] > 0.0)? Fr2_[iy]:0.0;
    m_fr2 +=(tmp1+tmp2)*dy;

    ReLambdaMax = (ReT_[iy] > ReLambdaMax) ? ReT_[iy]:ReLambdaMax;
    m_reT +=(ReT_[iy-1]+ReT_[iy])*dy;

    Row2Max = (Row2_[iy] > Row2Max) ? Row2_[iy]:Row2Max;
    m_row2 +=(Row2_[iy-1]+Row2_[iy])*dy;

    i_uu +=(uu_[iy-1]+uu_[iy])*dy;
    i_vv +=(vv_[iy-1]+vv_[iy])*dy;
    i_ww +=(ww_[iy-1]+ww_[iy])*dy;
    i_uv +=(uv_[iy-1]+uv_[iy])*dy;
    i_uw +=(uw_[iy-1]+uw_[iy])*dy;
    i_wv +=(wv_[iy-1]+wv_[iy])*dy;
    i_ss +=(ss_[iy-1]+ss_[iy])*dy;
    i_us +=(us_[iy-1]+us_[iy])*dy;
    i_vs +=(vs_[iy-1]+vs_[iy])*dy;
    i_ws +=(ws_[iy-1]+ws_[iy])*dy;
    i_ke +=(ke_[iy-1]+ke_[iy])*dy;
    i_tke +=(tke_[iy-1]+tke_[iy])*dy;
    i_pe +=(pe_[iy-1]+pe_[iy])*dy;
    
    i_fuu +=(fuu_[iy-1]+fuu_[iy])*dy;
    i_fvv +=(fvv_[iy-1]+fvv_[iy])*dy;
    i_fww +=(fww_[iy-1]+fww_[iy])*dy;
    i_ftke +=(ftke_[iy-1]+ftke_[iy])*dy;
    i_fus +=(fus_[iy-1]+fus_[iy])*dy;
    i_fvs +=(fvs_[iy-1]+fvs_[iy])*dy;
    i_fws +=(fws_[iy-1]+fws_[iy])*dy;
    i_fss +=(fss_[iy-1]+fss_[iy])*dy;
    i_fpe +=(fpe_[iy-1]+fpe_[iy])*dy;

//    i_rruu +=(rruu_[iy-1]+rruu_[iy])*dy;
//    i_rrww +=(rrww_[iy-1]+rrww_[iy])*dy;
//    i_rbtke +=(rbtke_[iy-1]+rbtke_[iy])*dy;
//    i_rrus +=(rrus_[iy-1]+rrus_[iy])*dy;
//    i_rbvs +=(rbvs_[iy-1]+rbvs_[iy])*dy;
//    i_rrws +=(rrws_[iy-1]+rrws_[iy])*dy;
//    i_rbpe +=(rbpe_[iy-1]+rbpe_[iy])*dy;
    i_uvdudy +=(uvdudy_[iy-1]+uvdudy_[iy])*dy;
    i_wvdwdy +=(wvdwdy_[iy-1]+wvdwdy_[iy])*dy;
    i_uvdsdy +=(uvdsdy_[iy-1]+uvdsdy_[iy])*dy;
    i_vvdsdy +=(vvdsdy_[iy-1]+vvdsdy_[iy])*dy;
    i_wvdsdy +=(wvdsdy_[iy-1]+wvdsdy_[iy])*dy;
    i_vsdudy +=(vsdudy_[iy-1]+vsdudy_[iy])*dy;
    i_vsdwdy +=(vsdwdy_[iy-1]+vsdwdy_[iy])*dy;
    i_dsdy +=(ds_dy(0,iy-1,0,0)+ds_dy(0,iy,0,0))*dy;

  }
  Measures[0] = i_ke*0.5;
  Measures[1] = i_tke*0.5;
  Measures[2] = i_pe*0.5;
  Measures[5] = m_reT*0.5/(u.b()-u.a());
  Measures[6] = ReLambdaMax;
  Measures[7] = m_fr2*0.5/(u.b()-u.a());
  Measures[8] = Fr2max;
  Measures[9] = Fr2min;
  Measures[10] = m_ri*0.5/(u.b()-u.a());
  Measures[11] = Rimax;
  Measures[12] = Rimin;
  Measures[13] = m_row2*0.5/(u.b()-u.a());
  Measures[14] = Row2Max;
  Measures[15] = i_uu*0.5;
  Measures[16] = i_vv*0.5;
  Measures[17] = i_ww*0.5;
  Measures[18] = i_uv*0.5;
  Measures[19] = i_wv*0.5;
  Measures[20] = i_uw*0.5;
  Measures[21] = i_ss*0.5;
  Measures[22] = i_us*0.5;
  Measures[23] = i_vs*0.5;
  Measures[24] = i_ws*0.5;
  Measures[25] = i_fuu*0.5;
  Measures[26] = i_fvv*0.5;
  Measures[27] = i_fww*0.5;
  Measures[28] = i_ftke*0.5;
  Measures[29] = i_fus*0.5;
  Measures[30] = i_fvs*0.5;
  Measures[31] = i_fws*0.5;
  Measures[32] = i_fss*0.5;
  Measures[33] = i_fpe*0.5;
  Measures[34] = -4.0*Omega*i_uw/(-2.0*i_uvdudy+1.0e-30);
  Measures[35] = 4.0*Omega*i_uw/(-2.0*i_wvdwdy+1.0e-30);
  Measures[36] = -g*i_vs/(-i_uvdudy-i_wvdwdy+1.0e-30);
  Measures[37] = -2.0*Omega*i_ws/(-i_uvdsdy-i_vsdudy+1.0e-30);
  Measures[38] = -g*i_ss/(-i_vvdsdy+1.0e-30);
  Measures[39] = 2.0*Omega*i_us/(-i_wvdsdy-i_vsdwdy+1.0e-30);
  Measures[40] = g*i_vs/(-2*g*kappa*i_dsdy+1.0e-30);

//  Measures[34] = i_rruu*0.5;
//  Measures[35] = i_rrww*0.5;
//  Measures[36] = i_rbtke*0.5;
//  Measures[37] = i_rrus*0.5;
//  Measures[38] = i_rbvs*0.5;
//  Measures[39] = i_rrws*0.5;
//  Measures[40] = i_rbpe*0.5;

  Real uu, ReTauTop, ReTauBot;
  uu = sqrt(square(du_dy(0,0,0,0))+square(dw_dy(0,0,0,0)));
  uu = sqrt(nu*fabs(uu));
  ReTauTop = uu*fabs(u.b()-u.a())*0.5/nu;
  Measures[3] = ReTauTop;
  uu = sqrt(square(du_dy(0,Ny-1,0,0))+square(dw_dy(0,Ny-1,0,0)));
  uu = sqrt(nu*fabs(uu));
  ReTauBot = uu*fabs(u.b()-u.a())*0.5/nu;
  Measures[4] = ReTauBot;
  
// recovery current state  
  u.makeState(uxzstate, uystate);
  q.makeState(qxzstate, qystate);
  s.makeState(sxzstate, systate);
   
  string filename = filebase;
  FILE *fpu;
  fpu = fopen(filename.c_str(), "w");
  if (fpu == NULL) {
    cerr << "FlowField::mean_profile can't open file " << endl;
    abort();
  }

  Real uc, sc, us, ss;
  uc = (u_(0,0,0,0)+u_(0,Ny-1,0,0))*0.5;
  sc = (s_(0,0,0,0)+s_(0,Ny-1,0,0))*0.5;
  us = u_(0,0,0,0)-uc;us = 1.0/(us+1.0e-40);
  ss = s_(0,0,0,0)-sc;ss = 1.0/(ss+1.0e-40);
  fprintf(fpu, 
  "#y u v w q s uu vv ww uv wv uw ss us vs ws Ri Fr2 N ReLambda TKE PE KE RoW2 alpha noru nors ");
  fprintf(fpu, 
  "fuu fvv fww ftke fus fvs fws fss fpe rruu rrww rbtke rrus rbvs rrws rbpe krho dis ");
  fprintf(fpu,
  "norkrho nordis div Ycg\n");
  for (iy=0; iy<Ny; iy++) {
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e ",
      u.y(iy)-u.b(), u_(0,iy,0,0), v_(0,iy,0,0), w_(0,iy,0,0), q_(0,iy,0,0), s_(0,iy,0,0));
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e ",
      uu_[iy], vv_[iy], ww_[iy], uv_[iy], wv_[iy], uw_[iy]);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e ", 
      ss_[iy], us_[iy], vs_[iy], ws_[iy]);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e %15.7e %15.7e %15.7e ", 
      Ri_[iy], Fr2_[iy], N_[iy], ReT_[iy], tke_[iy], pe_[iy], ke_[iy], Row2_[iy], alpha_[iy]);
    fprintf(fpu, "%15.7e %15.7e ", (u_(0,iy,0,0)-uc)*us, (s_(0,iy,0,0)-sc)*ss);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e %15.7e %15.7e %15.7e ", 
      fuu_[iy],fvv_[iy],fww_[iy],ftke_[iy],fus_[iy],fvs_[iy],fws_[iy],fss_[iy],fpe_[iy]);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e %15.7e ", 
      rruu_[iy],rrww_[iy],rbtke_[iy],rrus_[iy],rbvs_[iy],rrws_[iy],rbpe_[iy]);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e\n", 
      krho_[iy], dis_[iy], norkrho_[iy], nordis_[iy], div_[iy], Ycg_[iy]);
  }

  fclose(fpu);

}

void add_measures(const std::string& filebase, Real time, FlowField& u, FlowField& q, FlowField& s,
     Real nu, Real kappa, Real g, Real Omega) {
  assert(u.Nd()==3); assert(q.Nd()==1); assert(s.Nd()==1); 

// backup current state
  fieldstate uxzstate = u.xzstate();
  fieldstate uystate = u.ystate();
  fieldstate qxzstate = q.xzstate();
  fieldstate qystate = q.ystate();
  fieldstate sxzstate = s.xzstate();
  fieldstate systate = s.ystate();

  int ix, iy, iz;
  int Nx, Ny, Nz;
  Nx = u.Nx(); Ny=u.Ny(); Nz=u.Nz();

  FlowField dudy(Nx,Ny,Nz,3,u.Lx(),u.Lz(),u.a(),u.b());
  FlowField grads(Nx,Ny,Nz,3,s.Lx(),s.Lz(),s.a(),s.b());
  FlowField graddsdy(Nx,Ny,Nz,3,s.Lx(),s.Lz(),s.a(),s.b());
  
  grad(s, grads);
  grad(grads[2],graddsdy);
  ydiff(u, dudy);
  grads.makePhysical();
  graddsdy.makePhysical();
  dudy.makePhysical();
  u.makePhysical();
  
  Real HC[Ny], HD[Ny], VC[Ny], VD[Ny];
  
  for(iy = 0; iy<Ny; iy++) {
    HC[iy] = 0.0; HD[iy] = 0.0;
    VC[iy] = 0.0; VD[iy] = 0.0;
    for(ix = 0; ix<Nx; ix++) {
      for(iz = 0; iz<Nz; iz++) {
        HC[iy] +=square(u(ix,iy,iz,0)*graddsdy(ix,iy,iz,0)+u(ix,iy,iz,2)*graddsdy(ix,iy,iz,2));
        HD[iy] +=square(dudy(ix,iy,iz,0)*grads(ix,iy,iz,0)+dudy(ix,iy,iz,2)*grads(ix,iy,iz,2));
        VC[iy] +=square(u(ix,iy,iz,1)*graddsdy(ix,iy,iz,1));
        VD[iy] +=square(dudy(ix,iy,iz,1)*grads(ix,iy,iz,1));
      }
    }
    HC[iy] = sqrt(HC[iy]/((double)Nx*Nz));
    HD[iy] = sqrt(HD[iy]/((double)Nx*Nz));
    VC[iy] = sqrt(VC[iy]/((double)Nx*Nz));
    VD[iy] = sqrt(VD[iy]/((double)Nx*Nz));
  }
      
// recovery current state  
  u.makeState(uxzstate, uystate);
  q.makeState(qxzstate, qystate);
  s.makeState(sxzstate, systate);
  
  FILE *fp1;
  if (!fexist("add_measures.dat")) {
     fp1 = fopen("add_measures.dat", "w");
     if (fp1 == NULL) {
       cerr << "FlowField::add_measures can't open file " << endl;
       abort();
     }
     
     fprintf(fp1, "#time HCc HDc VCc VDc RVDc\n");
  
     fclose(fp1);
  }

  fp1 = fopen("add_measures.dat", "a");
  if (fp1 == NULL) {
    cerr << "FlowField::add_measures can't open file " << endl;
    abort();
  }
  
  iy = (Ny-1)/2;
  fprintf(fp1,"%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e\n",
    time, HC[iy], HD[iy], VC[iy], VD[iy], VD[iy]/(HC[iy]+HD[iy]+VC[iy]));
  fclose(fp1);
  
  string filename = filebase;
  FILE *fp2;
  fp2 = fopen(filename.c_str(), "w");
  if (fp2 == NULL) {
    cerr << "FlowField::add_measures can't open file " << endl;
    abort();
  }
  fprintf(fp2, "#y HC HD VC VD RVD\n");
  for(iy = 0; iy<Ny; iy++) {
    fprintf(fp2, "%15.7e %15.7e %15.7e %15.7e %15.7e %15.7e\n",
      u.y(iy)-u.b(), HC[iy], HD[iy], VC[iy], VD[iy], VD[iy]/(HC[iy]+HD[iy]+VC[iy]));
  }
  fclose(fp2);

}

void saveUSd(const std::string& filebase, FlowField& MProfile, Real nu, Real kappa)
{
  assert(MProfile.Nd() == 4);

  int j, Ny;
  Ny = MProfile.Ny();

  FlowField dMPdy(1,Ny,1,4,MProfile.Lx(),MProfile.Lz(),MProfile.a(),MProfile.b());

  ydiff(MProfile, dMPdy);
  dMPdy.makePhysical();
  MProfile.makePhysical();
  
  Real uu, yy, uu_b, yy_b;
  uu = dMPdy(0,0,0,0);
  uu = sqrt(nu*fabs(uu));
  uu_b = dMPdy(0,Ny-1,0,0);
  uu_b = sqrt(nu*fabs(uu_b));

  Real ww, yyw, ww_b, yyw_b;
  ww = dMPdy(0,0,0,2);
  ww = sqrt(nu*fabs(ww));
  ww_b = dMPdy(0,Ny-1,0,2);
  ww_b = sqrt(nu*fabs(ww_b));

  Real ss, yys, ss_b, yys_b;
  ss = dMPdy(0,0,0,3);
  ss = sqrt(kappa*fabs(ss));
  ss_b = dMPdy(0,Ny-1,0,3);
  ss_b = sqrt(kappa*fabs(ss_b));
  
  string filename = filebase;
  FILE *fpu;
  fpu = fopen(filename.c_str(), "w");
  if (fpu == NULL) {
    cerr << "FlowField::saveUSd can't open file " << endl;
    abort();
  }

  Real b, a, Utop, Ubot, Wtop, Wbot, Stop, Sbot;
  b = MProfile.b();
  a = MProfile.a();
  Utop = MProfile(0,0,0,0);
  Ubot = MProfile(0,Ny-1,0,0);
  Wtop = MProfile(0,0,0,2);
  Wbot = MProfile(0,Ny-1,0,2);
  Stop = MProfile(0,0,0,3);
  Sbot = MProfile(0,Ny-1,0,3);
  if (kappa <= 1.e-9) {
  fprintf(fpu, "#y u yb ub yw w ywb wb\n");
  for (j=0; j<=(Ny-1)/2; ++j) {
    yy = fabs((b-MProfile.y(j))*uu/nu);
    yy_b = fabs((MProfile.y(Ny-1-j)-a)*uu_b/nu);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e ",
       yy,fabs((Utop-MProfile(0,j,0,0))/(uu+1.0e-40)),yy_b,fabs((Ubot-MProfile(0,Ny-1-j,0,0))/(uu_b+1.0e-40)));
    yyw = fabs((b-MProfile.y(j))*ww/nu);
    yyw_b = fabs((MProfile.y(Ny-1-j)-a)*ww_b/nu);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e\n",
       yyw,fabs((Wtop-MProfile(0,j,0,2))/(ww+1.0e-40)),yyw_b,fabs((Wbot-MProfile(0,Ny-1-j,0,2))/(ww_b+1.0e-40)));
  }
  }
  else {
  fprintf(fpu, "#y u yb ub yw w ywb wb ys s ysb sb\n");
  for (j=0; j<=(Ny-1)/2; ++j) {
    yy = fabs((b-MProfile.y(j))*uu/nu);
    yy_b = fabs((MProfile.y(Ny-1-j)-a)*uu_b/nu);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e ",
       yy,fabs((Utop-MProfile(0,j,0,0))/(uu+1.0e-40)),yy_b,fabs((Ubot-MProfile(0,Ny-1-j,0,0))/(uu_b+1.0e-40)));
    yyw = fabs((b-MProfile.y(j))*ww/nu);
    yyw_b = fabs((MProfile.y(Ny-1-j)-a)*ww_b/nu);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e ",
       yyw,fabs((Wtop-MProfile(0,j,0,2))/(ww+1.0e-40)),yyw_b,fabs((Wbot-MProfile(0,Ny-1-j,0,2))/(ww_b+1.0e-40)));
    yys = fabs((b-MProfile.y(j))*ss/kappa);
    yys_b = fabs((MProfile.y(Ny-1-j)-a)*ss_b/kappa);
    fprintf(fpu, "%15.7e %15.7e %15.7e %15.7e\n",
       yys,fabs((Stop-MProfile(0,j,0,3))/(ss+1.0e-40)),yys_b,fabs((Sbot-MProfile(0,Ny-1-j,0,3))/(ss_b+1.0e-40)));
  }
  }
  
  fclose(fpu);  
  
}

void line_sequence(char *filename, Real time, FlowField& u, int id) {
  assert(u.Nd()==3); assert((id>=0 || id<=3));

// backup current state
  fieldstate uxzstate = u.xzstate();
  fieldstate uystate = u.ystate();

  int ix, iy, iz;
  int Nx, Ny, Nz;
  Nx = u.Nx();
  Ny = u.Ny();
  Nz = u.Nz();
  
  FILE *fpu;
  
  if (fexist(filename)==0) {
  fpu = fopen(filename, "w");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  fprintf(fpu, " NAN ");
  for (iy=0; iy<Ny; iy++) fprintf(fpu, "%15.7e", u.y(iy)-u.b());
  fprintf(fpu, "\n");
  fclose(fpu);
  }

  fpu = fopen(filename, "a");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  
  u.makePhysical();

  fprintf(fpu, "%15.7e", time);
  ix = Nx/2; iz = Nz/2;
  if (id < 3) {
  for (iy=0; iy<Ny; iy++) fprintf(fpu, "%15.7e", u(ix,iy,iz,id));
  }
  else {
  Real k;
  for (iy=0; iy<Ny; iy++) {
    k = 0.5*(square(u(ix,iy,iz,0))+square(u(ix,iy,iz,1))+square(u(ix,iy,iz,2)));
    fprintf(fpu, "%15.7e", k);
  }
  }
  fprintf(fpu, "\n");
  fclose(fpu);

  u.makeState(uxzstate, uystate);
  
}

void line_sequence(char *filename, Real time, FlowField& s) {

// backup current state
  fieldstate sxzstate = s.xzstate();
  fieldstate systate = s.ystate();

  int ix, iy, iz;
  int Nx, Ny, Nz;
  Nx = s.Nx();
  Ny = s.Ny();
  Nz = s.Nz();
  
  FILE *fpu;
  
  if (fexist(filename)==0) {
  fpu = fopen(filename, "w");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  fprintf(fpu, " NAN ");
  for (iy=0; iy<Ny; iy++) fprintf(fpu, "%15.7e", s.y(iy)-s.b());
  fprintf(fpu, "\n");
  fclose(fpu);
  }

  fpu = fopen(filename, "a");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  
  s.makePhysical();

  fprintf(fpu, "%15.7e", time);
  ix = Nx/2; iz = Nz/2;
  for (iy=0; iy<Ny; iy++) fprintf(fpu, "%15.7e", s(ix,iy,iz,0));
  fprintf(fpu, "\n");
  fclose(fpu);

  s.makeState(sxzstate, systate);
  
}

void line_sequence_LRi(char *filename, Real time, FlowField& LRi) {

  int iy;
  int Ny;
  Ny = LRi.Ny();
  
  FILE *fpu;
  
  if (fexist(filename)==0) {
  fpu = fopen(filename, "w");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  fprintf(fpu, " NAN ");
  for (iy=0; iy<Ny; iy++) fprintf(fpu, "%15.7e", LRi.y(iy)-LRi.b());
  fprintf(fpu, "\n");
  fclose(fpu);
  }

  fpu = fopen(filename, "a");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  
  LRi.makePhysical();

  fprintf(fpu, "%15.7e", time);
  for (iy=0; iy<Ny; iy++) fprintf(fpu, "%15.7e", LRi(0,iy,0,0));
  fprintf(fpu, "\n");
  fclose(fpu);
  
}

void line_sequence(char *filename, Real time, FlowField& s, bool p) {
  assert(p);
  
// backup current state
  fieldstate sxzstate = s.xzstate();
  fieldstate systate = s.ystate();

  int ix, iy, iz;
  int Nx, Ny, Nz;
  Nx = s.Nx();
  Ny = s.Ny();
  Nz = s.Nz();
  
  FILE *fpu;
  
  if (fexist(filename)==0) {
  fpu = fopen(filename, "w");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  fprintf(fpu, " NAN ");
  for (iy=0; iy<Ny; iy++) fprintf(fpu, "%15.7e", s.y(iy)-s.b());
  fprintf(fpu, "\n");
  fclose(fpu);
  }

  fpu = fopen(filename, "a");
  if (fpu == NULL) {
    cerr << "FlowField::line_sequence can't open file " << endl;
    abort();
  }
  
  s.makePhysical();

  fprintf(fpu, "%15.7e", time);
  ix = Nx/2; iz = Nz/2;
  Real ytop = s.y(0); 
  Real ybot = s.y(Ny-1);
  Real dy = ytop-ybot;
  Real stop = s(ix,0,iz,0);
  Real sbot = s(ix,Ny-1,iz,0);
  Real ftop = stop/dy;
  Real fbot = sbot/dy;
  Real sp;
  for (iy=0; iy<Ny; iy++) {
    sp = s.y(iy);
    sp = s(ix,iy,iz,0) - ((sp-ybot)*ftop-(sp-ytop)*fbot);
    fprintf(fpu, "%15.7e", sp);
  }
  fprintf(fpu, "\n");
  fclose(fpu);

  s.makeState(sxzstate, systate);
  
}

void AscSave(const std::string& filebase, FlowField& u) {

 // backup current state
  fieldstate uxzstate = u.xzstate();
  fieldstate uystate = u.ystate();

  int ix, iy, iz;
  int Nx, Ny, Nz;
  Nx = u.Nx(); Ny = u.Ny(); Nz = u.Nz();

  string filename = filebase;
  FILE *fpu;
  fpu = fopen(filename.c_str(), "w");
  if (fpu == NULL) {
    cerr << "FlowField::AscSave can't open file " << endl;
    abort();
  }

  u.makePhysical();
  iy = 0;
  for(ix = 0; ix<Nx; ix++) {
    for(iz = 0; iz<Nz; iz++) {
      fprintf(fpu, "%14.7e %14.7e %14.7e\n",u.x(ix), u.z(iz), u(ix,iy,iz,0));
    }
  }
  
  fclose(fpu);

// recovery current state  
  u.makeState(uxzstate, uystate);

}

void plot3DSave_2(const std::string& filebase, FlowField& dn, FlowField& un, 
            ChebyCoeff& Ubase, Real Re, Real Time) {
  assert(un.Nd()==3); assert(dn.Nd()==1);
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string filename = filebase;

  FILE *fp;
  fp = fopen(filename.c_str(), "wb");
  if (fp == NULL) {
    cerr << "FlowField::plot3DSave_2 : can't open file " << filename << endl;
    abort();
  }

// backup current state
  fieldstate dnxzstate = dn.xzstate();
  fieldstate dnystate = dn.ystate();
  fieldstate unxzstate = un.xzstate();
  fieldstate unystate = un.ystate();

  // get real profiles
  dn.makePhysical();
  un.makePhysical();

  int itmp[1];
  itmp[0]=un.Nx();
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=un.Ny();
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=un.Nz();
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  
  int i, j, k;
  float tmp[4];

  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Re;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Time;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  
  int Nx, Ny, Nz;
  Nx = un.Nx();
  Ny = un.Ny();
  Nz = un.Nz();
  
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = dn(i,j,k,0);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  fieldstate Us = Ubase.state();
  Ubase.makePhysical();
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = un(i,j,k,0)+Ubase[j];
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = un(i,j,k,1);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = un(i,j,k,2);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[1] = un(i,j,k,0)+Ubase[j];
            tmp[2] = un(i,j,k,1);
            tmp[3] = un(i,j,k,2);
            tmp[0] = (tmp[1]*tmp[1]+tmp[2]*tmp[2]+tmp[3]*tmp[3])*0.5;
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  Ubase.makeState(Us);

  fclose(fp);  

// recovery current state  
  dn.makeState(dnxzstate, dnystate);
  un.makeState(unxzstate, unystate);
  
}

void plot3DSave_2(const std::string& filebase, FlowField& dn, FlowField& un, 
            Real Re, Real Time) {
  assert(un.Nd()==3); assert(dn.Nd()==1);
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string filename = filebase;

  FILE *fp;
  fp = fopen(filename.c_str(), "wb");
  if (fp == NULL) {
    cerr << "FlowField::plot3DSave_2 : can't open file " << filename << endl;
    abort();
  }

// backup current state
  fieldstate dnxzstate = dn.xzstate();
  fieldstate dnystate = dn.ystate();
  fieldstate unxzstate = un.xzstate();
  fieldstate unystate = un.ystate();

  // get real profiles
  dn.makePhysical();
  un.makePhysical();

  int itmp[1];
  itmp[0]=un.Nx();
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=un.Ny();
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  itmp[0]=un.Nz();
  itmp[0] = BE(itmp[0]);
  fwrite(itmp, sizeof(int), 1, fp);
  
  int i, j, k;
  float tmp[4];

  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = 0.0;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Re;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  tmp[0] = Time;
  tmp[0] = BE(tmp[0]);
  fwrite(tmp, sizeof(float), 1, fp);
  
  int Nx, Ny, Nz;
  Nx = un.Nx();
  Ny = un.Ny();
  Nz = un.Nz();
  
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = dn(i,j,k,0);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = un(i,j,k,0);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = un(i,j,k,1);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[0] = un(i,j,k,2);
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }
  for (k=0; k<Nz; ++k) {
    for (j=0; j<Ny; ++j) {
        for (i=0; i<Nx; ++i) {
            tmp[1] = un(i,j,k,0);
            tmp[2] = un(i,j,k,1);
            tmp[3] = un(i,j,k,2);
            tmp[0] = (tmp[1]*tmp[1]+tmp[2]*tmp[2]+tmp[3]*tmp[3])*0.5;
            tmp[0] = BE(tmp[0]);
            fwrite(tmp, sizeof(float), 1, fp);
        }
    }
  }

  fclose(fp);  

// recovery current state  
  dn.makeState(dnxzstate, dnystate);
  un.makeState(unxzstate, unystate);
  
}

// save for tecplot 10
void tecplot10Save(const std::string& filebase, FlowField& dn, FlowField& un, Real time) {
  assert(un.Nd()==3); assert(dn.Nd()==1);

  static int zonenumber = 1;
  
  /************************************
  string filename; 
  if (filebase.length() > 
      3 && filebase.substr(filebase.length()-3, 3) == ".xyz")
    filename = filebase;
  else
    filename += string(".xyz"); // "flow field"
  *******************************/
  string fname = appendSuffix(filebase, ".plt");
  char filename[32];
  strcpy(filename, fname.c_str());
  
// backup current state
  fieldstate dnxzstate = dn.xzstate();
  fieldstate dnystate = dn.ystate();
  fieldstate unxzstate = un.xzstate();
  fieldstate unystate = un.ystate();

// get real velocities 
  un.makePhysical();
  dn.makePhysical();

  int II,III,Debug,DIsDouble,VIsDouble,IMax,JMax,KMax;
  
  int Nx_, Ny_, Nz_;
  Nx_ = un.Nx();
  Ny_ = un.Ny();
  Nz_ = un.Nz();
  
  Debug = 0;
  DIsDouble = 0;
  VIsDouble = 0;
  IMax = Nx_;
  JMax = Ny_;
  KMax = Nz_;
  III = IMax*JMax*KMax;

  II = TECINI("Velocity Dataset","X Y Z U V W S",&filename[0],".",&Debug,&VIsDouble);

  II = TECZNE("Simple Zone",&IMax,&JMax,&KMax,"BLOCK",NULL);
  
  int i, j, k;
  float tmp[Nz_][Ny_][Nx_];
  float tmp1;

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = un.x(i); tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = un.y(j)-un.b(); tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp1 = un.z(k); tmp[k][j][i] = tmp1;
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = un(i,j,k,0);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = un(i,j,k,1);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);
  
  for (k=0; k<Nz_; ++k) {
    for (j=0; j<Ny_; ++j) {
        for (i=0; i<Nx_; ++i) {
            tmp[k][j][i] = un(i,j,k,2);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  for (i=0; i<Nx_; ++i) {
    for (j=0; j<Ny_; ++j) {
        for (k=0; k<Nz_; ++k) {
            tmp[k][j][i] = dn(i,j,k,0);
        }
    }
  }
  II = TECDAT(&III,tmp,&DIsDouble);

  double xpos,ypos,fontheight,boxmargin,boxlinethickness,angle,linespacing;
  int poscoordmode,attachtozone,zone,font,fontheightunits;
  int boxtype,boxcolor,boxfillcolor,anchor,textcolor,scope;
  char text[32];

  xpos=70.;
  ypos=2.;
  poscoordmode=1;
  attachtozone=1;
  zone=zonenumber; //zonenumber++;
  font=1;
  fontheightunits=1;
  fontheight=3.;
  boxtype=0;
  boxmargin=5.;
  boxlinethickness=0.5;
  boxcolor=3;
  boxfillcolor=7;
  angle=0.;
  anchor=0;
  linespacing=1.5;
  textcolor=0;
  scope=1;

  sprintf(text, "Time=%10.3f", time);
  II = TECTXT(&xpos,&ypos,&poscoordmode,&attachtozone,&zone,&font,&fontheightunits,
        &fontheight,&boxtype,&boxmargin,&boxlinethickness,&boxcolor,&boxfillcolor,
        &angle,&anchor,&linespacing,&textcolor,&scope,&text[0],NULL);
  
  II = TECEND();
  
// recovery current state  
  dn.makeState(dnxzstate, dnystate);
  un.makeState(unxzstate, unystate);

}

void random1d(FlowField& sn, int positive) {
  assert(sn.Nd()==1);

  int i, j, k, Nx, Ny, Nz;
  Nx = sn.Nx();
  Ny = sn.Ny();
  Nz = sn.Nz();

  Real min, max;
  min = -0.5; max = 0.5;
  if (positive == 1) {min = 0.0; max = 1.0;}
  
  //Real tmp;
  
  sn.makePhysical();
  for (i=0; i<Nx; ++i) {
    for (j=0; j<Ny; ++j) {
        for (k=0; k<Nz; ++k) {
        sn(i,j,k,0) = randomReal(min,max);

        /*tmp = (sn.x(i)-1)*(sn.x(i)-1);
        tmp += sn.y(j)*sn.y(j);
        tmp += (sn.z(k)-1)*(sn.z(k)-1);
        sn(i,j,k,0) = exp(-tmp*4.0);*/

        // for Burger's equation - shock wave solution
        //sn(i,j,k,0) = 0.0;
        //if (sn.y(j) <= 0.0) sn(i,j,k,0) = 0.25;
            
        //sn(i,j,k,0) = sin(pi*sn.y(j))*sin(pi*sn.x(i))*sin(pi*sn.z(k)); // DbDt
        //sn(i,j,k,0) = cos(0.25*pi*(sn.y(j)+1.0))*sin(pi*sn.x(i))*sin(pi*sn.z(k)); // NbDt
        //sn(i,j,k,0) = sin(0.25*pi*(sn.y(j)+1.0))*sin(pi*sn.x(i))*sin(pi*sn.z(k)); // NbDt
        //sn(i,j,k,0)= 2.0*kappa*sin(pi*sn.y(j))/(1.5+cos(pi*sn.y(j)));
        }
    }
  }
  sn.makeSpectral();

}

void random3d(FlowField& un, int positive) {
  assert(un.Nd()==3);

  int i, j, k, Nx, Ny, Nz;
  Nx = un.Nx();
  Ny = un.Ny();
  Nz = un.Nz();

  Real min, max;
  min = -0.5; max = 0.5;
  if (positive == 1) {min = 0.0; max = 1.0;}
  
  un.makePhysical();
  for (i=0; i<Nx; ++i) {
    for (j=0; j<Ny; ++j) {
        for (k=0; k<Nz; ++k) {
        un(i,j,k,0) = randomReal(min,max);
        un(i,j,k,1) = randomReal(min,max);
        un(i,j,k,2) = randomReal(min,max);
        }
    }
  }
  un.makeSpectral();

}

int fexist(char *filename) {
  FILE* fp = NULL;
  fp = fopen(filename, "rb" );
  if( fp != NULL )
    {
    fclose( fp );
    return 1;
    }

  return 0 ;
}

void scalar_ini(FlowField& sn,  BoundaryC BC, Real bottom, Real top) {
  assert(sn.Nd()==1);

  int i, j, k, Nx, Ny, Nz;
  Real a, b;
  Nx = sn.Nx();
  Ny = sn.Ny();
  Nz = sn.Nz();
  a = sn.a();
  b = sn.b();

  Real min, max;
  min = 0.0; max = 1.0;

  sn.makePhysical();
  
  if (BC == DbDt) {
    Real c1, c2;
    c1 = (top-bottom)/(b-a);
    c2 = (b*bottom-a*top)/(b-a);
    for (i=0; i<Nx; ++i) {
        for (j=0; j<Ny; ++j) {
            for (k=0; k<Nz; ++k) {
                sn(i,j,k,0) = sn.y(j)*c1+c2;
            }
        }
    }
  }

  Real factor = pi*0.25;
  if (BC == DbNt) {
  
    if (fabs(top) <= 1.e-6) {
    for (i=0; i<Nx; ++i) {
        for (j=0; j<Ny; ++j) {
            for (k=0; k<Nz; ++k) {
                //sn(i,j,k,0) = bottom*sin(factor*(sn.y(j)/b+1.0))+bottom;
                sn(i,j,k,0) = 0.001*sin(factor*(sn.y(j)/b+1.0))+bottom;
            }
        }
    }
    }
    else {
    for (i=0; i<Nx; ++i) {
        for (j=0; j<Ny; ++j) {
            for (k=0; k<Nz; ++k) {
                sn(i,j,k,0) = top*(sn.y(j)/b+1.0)+bottom;
            }
        }
    }
    }
    
  }

  if (BC == NbDt) {

    Real ry, turn, add;
    turn = 0.2;
    add = 0.05;
    for (i=0; i<Nx; ++i) {
        for (j=0; j<Ny; ++j) {
            for (k=0; k<Nz; ++k) {
                //sn(i,j,k,0) = -top*sin(factor*(sn.y(j)/b-1.0))+top;
                //sn(i,j,k,0) = -0.001*sin(factor*(sn.y(j)/b-1.0))+top;
                //sn(i,j,k,0) = bottom*(sn.y(j)/b-1.0)+top;
                ry = (b-sn.y(j))/(b-a);
                sn(i,j,k,0) = (ry > turn)? top+add+bottom*(turn-ry):top+ry/turn*add;
            }
        }
    }
    
  }
  
  if (BC == NbNt) {
    for (i=0; i<Nx; ++i) {
        for (j=0; j<Ny; ++j) {
            for (k=0; k<Nz; ++k) {
            sn(i,j,k,0) = randomReal(min,max);
            }
        }
    }
  }
  
  sn.makeSpectral();

}

// initial a classical plane Couette-Poiseuille flow
// u = 0.5*dP/dx/nu*(y^2-1)+0.5*Usurf(y+1), -1<=y<=1
void velocity_ini(FlowField& un, Real Usurf, Real dPdx_nu) {
  assert(un.Nd()==3);

  int i, j, k, Nx, Ny, Nz;
  Nx = un.Nx();
  Ny = un.Ny();
  Nz = un.Nz();

  un.makePhysical();
  
  Real factor1 = Usurf*0.5;
  Real factor2 = dPdx_nu*0.5;
  for (i=0; i<Nx; ++i) {
    for (j=0; j<Ny; ++j) {
        for (k=0; k<Nz; ++k) {
            un(i,j,k,0) = factor1*un.y(j)+factor1+factor2*square(un.y(j))-factor2;
        }
    }
  }
  un.makeSpectral();

}

void heat_error(Real time, Real dt, Real kappa, FlowField& sn) {

  FILE *fp;
  fp = fopen("evolution_error.dat", "a");
  if (fp == NULL) {
    cerr << "FlowField::evolution can't open file " << endl;
    abort();
  }
  
  Real error = 0.0;
  Real U, exact;
  
  U = 0.25;
  
  int Ny = sn.Ny();
  Real uex[Ny], x[Ny];
  for (int j=0; j < Ny; ++j) x[j]=sn.y(Ny-1-j);
  burgers(uex, x, kappa/U, U*time, Ny);
  
  sn.makePhysical();
  for (int i=0; i<sn.Nx(); ++i) {
    for (int j=0; j<sn.Ny(); ++j) {
        for (int k=0; k<sn.Nz(); ++k) {
        exact = uex[Ny-1-j]*U;
        //exact = sin(pi*sn.y(j))*sin(pi*sn.x(i))*sin(pi*sn.z(k))*exp(-kappa*(3.0*pi*pi)*time); // DbDt
        //exact = cos(0.25*pi*(sn.y(j)+1.0))*sin(pi*sn.x(i))*sin(pi*sn.z(k))*exp(-kappa*(2.0625*pi*pi)*time); // NbDt
        //exact = sin(0.25*pi*(sn.y(j)+1.0))*sin(pi*sn.x(i))*sin(pi*sn.z(k))*exp(-kappa*(2.0625*pi*pi)*time); // DbNt
        //exact = 2.0*kappa*sin(pi*sn.y(j))*exp(-kappa*pi*pi*time)/(1.5+cos(pi*sn.y(j))*exp(-kappa*pi*pi*time));
        error +=(exact-sn(i,j,k,0))*(exact-sn(i,j,k,0));
        }
    }
  }
  //error = sqrt(error/((double)sn.Nx()*sn.Ny()*sn.Nz()));
  error = sqrt(error/((double)sn.Nx()*sn.Nz()));
  //error = sqrt(error);
  sn.makeSpectral();
  
  fprintf(fp, "%15.7e %15.7e %15.7e\n", time, dt, error);
  
  fclose(fp);

}  

/******************************************************************** $PAGE */
/* NAME       :derf                                                         */
/* FUNCTION   :Program estimates the error function of xarg                 */
/*             Source:   Abramowitz/Stegun, Hndbk of Math. Functions        */
/*                        Eq No. 7.1.26                                     */
/*             Accuracy: abs(exact-estimate) <= 1.5e-7                      */
/* INPUT      :                                                             */
/* OUTPUT     :-                                                            */
/* RETURN     :-                                                            */
/****************************************************************************/
Real derf (Real xarg)
{
  Real result;     
  Real x, t, z;
  const Real p =   0.32759110;
  const Real a1 =  0.254829592;
  const Real a2 = -0.284496736;
  const Real a3 =  1.421413741;
  const Real a4 = -1.453152027;
  const Real a5 =  1.061405429; 
  x = fabs(xarg);
  t = 1.0 / ( 1.0 + p * x );
  
  Real x2, t2, t3, t4, t5;
  x2 = x*x;
  t2 = t*t;
  t3 = t2*t;
  t4 = t3*t;
  t5 = t4*t;
  
  z =  exp( -x2 ) * (    a1 * t 
                             +  a2 * t2
                             +  a3 * t3
                             +  a4 * t4
                             +  a5 * t5 );
  result = 1.0 - z;
  if( xarg < 0.0 ) 
  {
    result = -result;
  }      
  return( result );
}  /* of derf */

/******************************************************************** $PAGE */
/* NAME       :burgers                                                      */
/* FUNCTION   :Routine calculates the solution to the non-linear burgers    */
/*             equation on the domain -1 to 1. Assumes the initial          */
/*             conditions are u(x<0) = 1, u(x>0) = 0  with compatible       */
/*             boundary conditions.                                         */
/* INPUT      : x        grid from -1 to 1                                  */
/*              xnu      viscosity                                          */
/*              time     time                                               */
/* OUTPUT     :-                                                            */
/* RETURN     :uex = burgers eqn solution on the grid                       */
/****************************************************************************/
void burgers (Real *uex, Real *x, Real xnu, Real time, int nx)
{
  
  int i;
  Real fac, fac1, fac2;
  Real erfac1, erfac2;

  if ( time <= 0.0 ) {
    printf("Error! Function Burger does only work for t > 0 \n");
    printf("Program terminated\n");
    printf("\a");
    exit(1);
  }

 
  // boundary conditions
  uex[0]      = 1.0;
  uex[nx - 1] = 0.0;

  // interior
  fac1 = sqrt( 0.25 * pi );
  for ( i = 1; i < nx - 1; i++ )
  {
    fac2   = sqrt( pi * time * xnu );
    erfac1 = ( x[i] - time ) * fac1 / fac2;
    erfac2 =  -x[i] * fac1 / fac2;
    erfac1 = 1.0 - derf(erfac1);
    erfac2 = 1.0 - derf(erfac2);
    fac    = ( 0.5 * ( x[i] - 0.5 * time ) / xnu );
    if( fac < 200.0 ) 
    {
      fac = exp(fac);
      uex[i] = erfac1 / ( erfac1 + fac * erfac2 );
    }
    else
    { 
      uex[i] = 0.0;
    }
  }
  return;
} /* of burgers */

} //namespace channelflow
